import unittest
from birdwatcher.frames import framecolor, framegray
from birdwatcher.plotting import imshow_frame


class TestImshowFrame(unittest.TestCase):

    def test_imshowcolor(self):
        frame = framecolor(height=480, width=640, color=(1,2,3))
        imshow_frame(frame)

    def test_imshowgray(self):
        frame = framegray(height=480, width=640)
        imshow_frame(frame)




"""This module provides objects and functions for coordinate data.
Coordinates are pixel positions (x,y) in a Frame. Coordinate Arrays are a
convenient way of storing output from image algorithms that determine if a
pixel is categorized as something (e.g. crossing some threshold).
Information is memory-mapped from disk, because data can easily become very
large and will not fit in RAM memory. Since the number of pixels per frame may
be variable, we use Ragged Arrays from the python library Darr to store them.
This is not the most disk-space efficient way of doing this (no compression),
but it is fast and the data can easily be read in any scientific computing
environement (Python, Matlab, R, Mathematica, etc.) Coordinate files can be
archived in compressed form (lzma) to save disk space.

"""

import os
import shutil
import tarfile
from contextlib import contextmanager
from pathlib import Path

import numpy as np
from darr import RaggedArray, delete_raggedarray, create_raggedarray

from ._version import get_versions
from .utils import tempdir
from .frames import frameiterator


__all__ = ['CoordinateArrays', 'create_coordarray', 
           'open_archivedcoordinatedata', 'extract_archivedcoordinatedata', 
           'delete_coordinatearray', 'move_coordinatearrays']


def _archive(rar):
    rar.archive(overwrite=True)
    delete_raggedarray(rar)


def _coordstoframe(coords, width, height, nchannels=None, dtype='uint8',
                   value=1):
    if nchannels is None:
        frame = np.zeros((height, width), dtype=dtype)
    else:
        frame = np.zeros((height, width, nchannels), dtype=dtype)
    frame[(coords[:, 1], coords[:, 0])] = value
    return frame


# fixme, should we allow for 3 values: (x, y, val)?
class CoordinateArrays(RaggedArray):
    """A disk-based data type to store frame coordinates of consecutive
    frames.

    Maximum for frame width and height is 65535.

    Parameters
    ----------
    path : str or pathlib.Path
        Path to disk-based coordinate array directory.
    accessmode : {'r', 'r+'}, default 'r'
       File access mode of the darr data. `r` means read-only, `r+` means
       read-write. `w` does not exist. To create new coordinate arrays,
       potentially overwriting an other one, use the `create_coordarray`
       functions.

    """

    def __init__(self, path, accessmode='r'):
        super().__init__(path=path, accessmode=accessmode)
        md = dict(self.metadata)
        self.framewidth = md['framewidth']
        self.frameheight = md['frameheight']

    def get_frame(self, frameno, nchannels=None, dtype='uint8', value=1):
        """Get a frame based on a sequence number in the coordinate array.

        Parameters
        ----------
        frameno : int
            The sequence number of the frame to get.
        nchannels : int, optional
            The number of color channels in the frame. Default None leads
            to no color dimension, just a 2D frame with gray values.
        dtype : numpy dtype, default='uint8'
            Dtype of the returned frame.
        value : int, default=1
            The value to set the present coordinates with.

        Returns
        -------
        Numpy array

        """
        return _coordstoframe(coords=self[frameno], width=self.framewidth,
                              height=self.frameheight, nchannels=nchannels,
                              dtype=dtype, value=value)

    @frameiterator
    def iter_frames(self, startframe=0, endframe=None, stepsize=1,
                    nchannels=None, dtype='uint8', value=1):
        """Iterate over coordinate array and produce frames.

        Parameters
        ----------
        startframe : int, optional
            If specified, start iteration at this frame number.
        endfrom : int, optional
            Frame number to end iteration at. Default is None, which is to
            the end.
        stepsize : int, optional
            Step sizes. Defaults to 1, but if you want to skip frames, you
            can use this parameter.
        nchannels : int, optional
            The number of color channels in the frame. Default is None which
            leads to no color dimension, just a 2D frame with gray values.
        dtype : numpy dtype, default='uint8'
            Dtype of the returned frame.
        value : int, default=1
            The value to set the present coordinates with.

        Yields
        ------
        Frames
            Iterator that produces video frames based on the coordinates.

        """
        for coords in self.iter_arrays(startindex=startframe,
                                       endindex=endframe, stepsize=stepsize):
            yield _coordstoframe(coords=coords, width=self.framewidth,
                                 height=self.frameheight, nchannels=nchannels,
                                 dtype=dtype, value=value)

    def tovideo(self, filepath, startframe=0, endframe=None, stepsize=1,
                framerate=None, crf=17, scale=None, format='mp4',
                codec='libx264', pixfmt='yuv420p', ffmpegpath='ffmpeg'):
        """Writes frames based on coordinate info to a video file.

        Parameters
        ----------
        filepath : str
            Name of the videofilepath that should be written to.
        startframe: int, optional
            If specified, start iteration at this frame number.
        endfrom : int, optional
            Frame number to end iteration at. Default is None, which is to
            the end.
        stepsize : int, optional
            Step sizes. Defaults to 1, but if you want to skip frames, you
            can use this parameter.
        framerate : int, optional
            framerate of video in frames per second. If None, will look for
            `avgframerate` in metadata.
        crf : int, default=17
            Value determines quality of video. The default 17 is high quality.
            Use 23 for good quality.
        scale : tuple, optional
            (width, height). The default (None) does not change width and
            height.
        format : str, default='mp4'
            ffmpeg video format.
        codec : str, default='libx264'
            ffmpeg video codec.
        pixfmt : str, default='yuv420p'
            ffmpeg pixel format.
        ffmpegpath : str or pathlib.Path, optional
            Path to ffmpeg executable. Default is `ffmpeg`, which means it
            should be in the system path.

        Notes
        -----
        See ffmpeg documentation for more information.

        """
        from .ffmpeg import arraytovideo
        if framerate is None:
            try:
                framerate = self.metadata['avgframerate']
            except KeyError:
                raise ValueError('Cannot find a frame rate, you need to '
                                 'provide one with the `framerate` parameter')
        arraytovideo(self.iter_frames(startframe=startframe,
                                      endframe=endframe,
                                      stepsize=stepsize, nchannels=3,
                                      value=255,
                                      dtype='uint8'),
                     filepath, framerate=framerate, scale=scale, crf=crf,
                     format=format, codec=codec, pixfmt=pixfmt,
                     ffmpegpath=ffmpegpath)

    def show(self, startframe=0, endframe=None, stepsize=1, framerate=None,
             draw_framenumbers=True):
        """Shows coordinates frames in a video window.

        Turns each coordinate array into a frame and then plays video.

        Parameters
        ----------
        startframe: int, optional
            If specified, start iteration at this frame number.
        endfrom : int, optional
            Frame number to end iteration at. Default is None, which is to
            the end.
        stepsize : int, optional
            Step sizes. Defaults to 1, but if you want to skip frames, you
            can use this parameter.
        framerate : int, optional
            framerate of video in frames per second. If None, will look for
            `avgframerate` in metadata.
        draw_framenumbers : bool, default=True
            Should I draw frame numbers yes or no?

        Returns
        -------

        """
        if framerate is None:
            try:
                framerate = self.metadata['avgframerate']
            except KeyError:
                raise ValueError('Cannot find a frame rate, you need to '
                                 'provide one with the `framerate` parameter')

        f = self.iter_frames(startframe=startframe, endframe=endframe,
                         stepsize=stepsize, nchannels=3, value=255,
                         dtype='uint8')
        if draw_framenumbers:
            f = f.draw_framenumbers()
        return f.show(framerate=framerate)

    def get_coordcount(self, startframeno=0, endframeno=None):
        """Get the number of coordinates present per frame.

        Parameters
        ----------
        startframeno : int, optional
            Defaults to the beginning of the coordinate array.
        endframeno : int, optional
            Defaults to the end of the coordinate array.

        Returns
        -------
        Numpy Array
            Sequence of numbers, each with a coordinate count.

        """
        coordgen = self.iter_arrays(startindex=startframeno,
                                    endindex=endframeno)
        return np.array([c.shape[0] for c in coordgen])

    def get_coordmean(self, startframeno=0, endframeno=None):
        """Get the mean of the coordinates per frame.

        Parameters
        ----------
        startframeno : int, optional
            Defaults to the beginning of the coordinate array.
        endframeno : int, optional
            Defaults to the end of the coordinate array.

        Returns
        -------
        Numpy Array
            Sequence of numbers, each with a coordinate mean.

        """
        coordgen = self.iter_arrays(startindex=startframeno,
                                    endindex=endframeno)
        return np.array([c.mean(0) if c.size>0 else (np.nan, np.nan)
                         for c in coordgen])

    def get_coordmedian(self, startframeno=0, endframeno=None):
        """Get the median of the coordinates per frame.

        Parameters
        ----------
        startframeno : int, optional
            Defaults to the beginning of the coordinate array.
        endframeno : int, optional
            Defaults to the end of the coordinate array.

        Returns
        -------
        Numpy Array
            Sequence of numbers, each with a coordinate median.

        """
        coordgen = self.iter_arrays(startindex=startframeno,
                                    endindex=endframeno)
        return np.array([np.median(c,0) if c.size>0 else (np.nan, np.nan)
                         for c in coordgen])


def create_coordarray(path, framewidth, frameheight, metadata=None,
                      overwrite=True):
    """Creates an empty Coordinate Arrays object.

    Parameters
    ----------
    path : str
        Path to disk-based coordinate array directory that should be written
        to.
    framewidth : int
        Width in pixels.
    frameheight : int
        Height in pixels.
    metadata : dict, optional
    overwrite : bool, default=True
        Overwrite existing CoordinateArrays or not.

    Returns
    -------
    CoordinateArrays

    """
    Path(path).parent.mkdir(parents=True, exist_ok=True)   
    if metadata is None:
        metadata = {}
    metadata.update({'framewidth': framewidth,
                     'frameheight': frameheight,
                     'birdwatcher_version': get_versions()['version']})
    coords = create_raggedarray(path, atom=(2,), metadata=metadata,
                                overwrite=overwrite, dtype='uint16')
    return CoordinateArrays(coords.path, accessmode='r+')


@ contextmanager
def open_archivedcoordinatedata(path, temppath=None):
    """A context manager that temporarily decompresses coordinate
    data to work with coordinate array.

    Parameters
    ----------
    path : str
        Path to the archive.

    Yields
    ------
    Context manager to work with temporarily uncompressed coordinate
    array.

    Examples
    --------
    >>> with open_archivedcoordinatedata('coords.darr.tar.xz') as coords:
            # do stuff with coordinate array

    """
    path = Path(path)
    if not path.suffix == '.xz':
        raise OSError(f'{path} does not seem to be archived coordinate data')

    with tempdir(dirname=temppath) as dirname:
        tar = tarfile.open(path)
        tar.extractall(path=dirname)
        tar.close()
        capath = list(dirname.glob('*'))[0]
        yield CoordinateArrays(capath)


def extract_archivedcoordinatedata(path):
    """An archived coordinate array (.tar.xz) is decompressed.

    Parameters
    ----------
    path : str
        Path to the archived coordinate array.

    Returns
    -------
    CoordinateArrays

    """
    path = Path(path)
    if not path.suffix == '.xz':
        raise OSError(f'{path} does not seem to be archived coordinate data')

    tar = tarfile.open(path)
    tar.extractall(path=path.parent)
    tar.close()
    
    while path.suffix in {'.tar', '.xz'}: # remove extensions
        path = path.with_suffix('')

    return CoordinateArrays(path)


delete_coordinatearray = delete_raggedarray


def move_coordinatearrays(sourcedirpath, targetdirpath):
    """Move coordinate / darr data hierarchically out of a source dir and
    move it to a target dir, keeping the hierarchy intact.

    This is handy when you created a zillion coordinate / darr inside some
    directory hierarchy of input data, and you want to separate things.

    Parameters
    ----------
    sourcedirpath : str or Path
    targetdirpath : str or Path
        The top-level directory to which everything is moved. If it doesn't
        exist it will be created.

    """
    tdir = Path(targetdirpath)
    for root, dirs, files in os.walk(sourcedirpath):
        for dname in dirs:
            if dname.endswith('.darr'):
                d = Path(dname)
                newdir = tdir / Path(root).relative_to(sourcedirpath)
                Path(newdir).mkdir(parents=True, exist_ok=True)
                shutil.move(f"{root}/{d}", str(newdir))


"""This module contains classes for generating image frames and general
processing functionality such as measurement, drawing of labels/text and
saving as video files.

Many methods mirror functions from OpenCV. Docstrings are provided but
it is a good idea to look at OpenCV's documentation and examples if you want
to understand the parameters in more depth.

"""

from functools import wraps
from pathlib import Path

import numpy as np
import cv2 as cv

from .utils import peek_iterable


__all__ = ['Frames', 'FramesColor', 'FramesGray', 'FramesNoise', 'framecolor',
           'framegray', 'framenoise']


def _check_writable(frame):
    if not frame.flags.writeable:
        return frame.copy()
    else:
        return frame


def frameiterator(func):
    @wraps(func)
    def wrapper(*args, **kwargs):
        processingdata = []
        self = args[0]
        if hasattr(self, 'get_info'):
            processingdata = self.get_info().get('processingdata') or []
        processingdata.append({'methodname': func.__name__,
                               'methodargs': [str(arg) for arg in args],
                               'methodkwargs': dict((str(key),str(item))
                                                    for (key, item)
                                                    in kwargs.items())})
        return Frames(func(*args, **kwargs), processingdata=processingdata)
    return wrapper

class Frames:
    """An iterator of video frames with useful methods.

    This is a main base class in Birdwatcher, as many functions and
    methods return this type and take it as input. Many methods of `Frames`
    objects return new `Frames` objects, but some of them generate final
    output, such as a video file or a measurement.

    Parameters
    ----------
    frames : iterable
        This can be anything that is iterable and that produces image frames.
        A numpy array, a VideoFileStream or another Frames object.

    Examples
    --------
    >>> import birdwatcher as bw
    >>> frames = bw.FramesNoise(250, height=720, width=1280)
    >>> frames = frames.draw_framenumbers()
    >>> frames.tovideo('noisewithframenumbers.mp4', framerate=25)
    >>> # next example based on input from video file
    >>> vfs = bw.VideoFileStream('zebrafinchrecording.mp4')
    >>> frames = vfs.iter_frames() # create Frames object
    >>> # more concise expression
    >>> frames.blur(ksize=(3,3)).togray().tovideo('zf_blurgray.mp4')

    """

    def __init__(self, frames, processingdata=None):

        first, frames = peek_iterable(frames)

        frameheight, framewidth, *nchannels = first.shape
        if nchannels == []:
            nchannels = [1]
        self._frames = frames
        self._frameheight = frameheight
        self._framewidth = framewidth
        self._nchannels = nchannels[0]
        self._index = 0
        self._dtype = first.dtype.name
        self.processingdata = processingdata

    def __iter__(self):
        return self

    def __next__(self):
        return next(self._frames)

    @property
    def frameheight(self):
        return self._frameheight

    @property
    def framewidth(self):
        return self._framewidth

    @property
    def nchannels(self):
        return self._nchannels

    @property
    def dtype(self):
        return self._dtype

    def get_info(self):
        return {'classname': self.__class__.__name__,
                'framewidth': self.framewidth,
                'frameheight': self.frameheight,
                'processingdata': self.processingdata}
    
    def peek_frame(self):
        """Returns first frame without removing it.
        
        Returns
        -------
        numpy ndarray
            The first frame.
        """
        firstframe, self._frames = peek_iterable(self._frames)
        return firstframe
        
    def tovideo(self, filepath, framerate, crf=23, scale=None, format='mp4',
                codec='libx264', pixfmt='yuv420p', ffmpegpath='ffmpeg'):
        """Writes frames to video file.

        Parameters
        ----------
        filepath : str
            Name of the videofilepath that should be written to.
        framerate : int
            framerate of video in frames per second.
        crf : int, default=23
            Value determines quality of video. The default 23 is good quality.
            Use 17 for high quality.
        scale : tuple, optional
            (width, height). The default (None) does not change width and    
            height.
        format : str, default='mp4'
            ffmpeg video format.
        codec : str, default='libx264'
            ffmpeg video codec.
        pixfmt : str, default='yuv420p'
            ffmpeg pixel format.
        ffmpegpath : str or pathlib.Path, optional
            Path to ffmpeg executable. Default is `ffmpeg`, which means it
            should be in the system path.

        Notes
        -----
        See ffmpeg documentation for more information.

        """
        from .ffmpeg import arraytovideo
        from .video import VideoFileStream
        filepath = arraytovideo(frames=self, filepath=filepath,
                                framerate=framerate, crf=crf, scale=scale,
                                format=format, codec=codec, pixfmt=pixfmt,
                                ffmpegpath=ffmpegpath)
        return VideoFileStream(filepath)

    @frameiterator
    def blur(self, ksize, anchor=(-1,-1), borderType=cv.BORDER_DEFAULT):
        """Blurs frames using the normalized box filter.

        Parameters
        ----------
        ksize : (int, int)
            Kernel size. Tuple of integers.
        anchor : (int, int), optional
            Anchor point. Default value (-1,-1) means that the anchor is at
            the kernel center.
        borderType : int, default=cv.BORDER_DEFAULT
            Border mode used to extrapolate pixels outside of the image.

        Yields
        ------
        Frames
            Iterator that generates blurred frames.

        Examples
        --------
        >>> import birdwatcher as bw
        >>> frames = bw.FramesNoise(250, height=720, width=1280)
        >>> frames = frames.blur(ksize=(10,10))
        >>> frames.tovideo('noiseblurred.mp4', framerate=25)

        """
        for frame in self._frames:
            yield cv.blur(frame, ksize=ksize, anchor=anchor,
                          borderType=borderType)

    @frameiterator
    def edge_detection(self, minval=80, maxval=150):
        """Finds edges (boundaries) in frames.

        Only works on gray frames! Blur frames before applying edge detection
        for optimal results. Edges are defined by sudden changes in pixel
        intensity.

        Parameters
        ----------
        minval : str, optional
            Lower threshold for finding smaller edges.
        maxval : str, optional
            Higher threshold to determine segments of strong edges.

        Yields
        ------
        Frames
            Iterator that generates frames with edges.

        """
        for frame in self._frames:
            yield cv.Canny(frame, minval, maxval)

    # FIXME multiple circles per frame?
    @frameiterator
    def draw_circles(self, centers, radius=6, color=(255, 100, 0),
                     thickness=2, linetype=cv.LINE_AA, shift=0):
        """Draws circles on frames.

        Centers should be an iterable that has a length that corresponds to
        the number of frames.

        Parameters
        ----------
        centers : iterable
            Iterable that generate center coordinates (x, y) of the circles
        radius : int, default=6
            Radius of circle.
        color : tuple of ints, optional
            Color of circle (BGR). The default (255, 100, 0) color is blue.
        thickness : int, default=2
            Line thickness.
        linetype : int, default=cv2.LINE_AA
            OpenCV line type of circle boundary.
        shift : int, default=0
            Number of fractional bits in the coordinates of the center and in
            the radius value.

        Yields
        ------
        Frames
            Iterator that generates frames with circles.

        """
        for frame, center in zip(self._frames, centers):
            frame = _check_writable(frame)
            center = np.asanyarray(center)
            if not np.isnan(center).any():
                (x, y) = center.astype('int16')
                yield cv.circle(frame, center=(x, y), radius=radius,
                                color=color,
                                thickness=thickness, lineType=linetype,
                                shift=shift)
            else:
                yield frame

    @frameiterator
    def draw_rectangles(self, points, color=(255, 100, 0), thickness=2):
        """Draws rectangles on frames.

        Points should be an iterable that has a length that corresponds to
        the number of frames.

        Parameters
        ----------
        points : iterable
            Iterable that generates sequences of rectangle corners ((x1, y1),
            (x2, y2)) per frame, where the coordinates specify opposite
            corners (e.g. top-left and bottom-right).
        color : tuple of ints, optional
            Color of rectangle (BGR). The default (255, 100, 0) color is blue.
        thickness : int, default=2
            Line thickness.

        Yields
        ------
        Frames
            Iterator that generates frames with rectangles.

        """
        for frame, framepoints in zip(self._frames, points):
            frame = _check_writable(frame)
            framepoints = np.asanyarray(framepoints)
            if not np.isnan(framepoints).any():
                (pt1, pt2) = framepoints.astype('int16')
                yield cv.rectangle(frame, pt1=pt1, pt2=pt2, color=color, thickness=thickness)
            else:
                yield frame

    @frameiterator
    def togray(self):
        """Converts color frames to gray frames using OpenCV.

        Yields
        ------
        Frames
            Iterator that generates gray frames.

        """
        for frame in self._frames:
            if frame.ndim == 3:
                yield cv.cvtColor(frame, cv.COLOR_BGR2GRAY)
            else:
                yield frame

    @frameiterator
    def tocolor(self):
        """Converts gray frames to color frames using OpenCV.

        Yields
        ------
        Frames
            Iterator that generates color frames.

        """
        for frame in self._frames:
            if frame.ndim == 2:
                yield cv.cvtColor(frame, cv.COLOR_GRAY2BGR)
            else:
                yield frame

    @frameiterator
    def draw_framenumbers(self, startat=0, org=(2, 25),
                          fontface=cv.FONT_HERSHEY_SIMPLEX,
                          fontscale=1, color=(200, 200, 200), thickness=2,
                          linetype=cv.LINE_AA):
        """Draws the frame number on frames.

        Parameters
        ----------
        startat : int, optional
            The number to start counting at.
        org : (int, int), optional
            A tuple of ints (horizontal coordinate value, vertical coordinate
            value) indicates where to draw the framenumbers. The default (2,
            25) draws numbers in the top left corner of the image.
        fontface : OpenCV font type, default=cv.FONT_HERSHEY_SIMPLEX
        fontscale : float, optional
            Font scale factor that is multiplied by the font-specific base
            size.
        color : (int, int, int), optional
            Font color (BGR). The default (200, 200, 200) color is gray.
        thickness : int, default=2
            Line thickness.
        linetype : int, default=cv2.LINE_AA
            OpenCV line type.

        Yields
        ------
        Frames
            Iterator that generates frames with frame numbers.

        """
        for frameno, frame in enumerate(self._frames):
            frame = _check_writable(frame)
            yield cv.putText(frame, str(frameno + startat), org=org,
                             fontFace=fontface, fontScale=fontscale,
                             color=color, thickness=thickness,
                             lineType=linetype)

    @frameiterator
    def draw_text(self, textiterator, org=(2, 25),
                  fontface=cv.FONT_HERSHEY_SIMPLEX, fontscale=1,
                  color=(200, 200, 200), thickness=2, linetype=cv.LINE_AA):
        """Draws text on frames.

            Parameters
            ----------
            textiterator : iterable
                Something that you can iterate over and that produces text
                for each frame.
            org : (int, int), optional
                A tuple of ints (horizontal coordinate value, vertical
                coordinate value) indicates where to draw the text. The
                default (2, 25) draws text in the top left corner of the
                image.
            fontface : OpenCV font type, default=cv.FONT_HERSHEY_SIMPLEX
            fontscale : float, optional
                Font scale factor that is multiplied by the font-specific base
                size.
            color : (int, int, int), optional
                Font color (BGR). The default color (200, 200, 200) is gray.
            thickness : int, default=2
                Line thickness.
            linetype : int, default=cv2.LINE_AA
                OpenCV line type.

            Yields
            ------
            Frames
                Iterator that generates frames with text.

        """
        for frame, text in zip(self._frames, textiterator):
            frame = _check_writable(frame)
            yield cv.putText(frame, str(text), org=org,
                             fontFace=fontface, fontScale=fontscale,
                             color=color, thickness=thickness,
                             lineType=linetype)
        
    def save_nonzero(self, filepath, metadata, ignore_firstnframes=10, 
                     overwrite=True):
        """Save nonzero pixel coordinates (i.e. foreground) as Coordinate 
        Arrays object.

        Parameters
        ----------
        filepath : str
            Name of the filepath that should be written to.
        metadata : dict, optional
        ignore_firstnframes : int, default=10
            Do not provide coordinates for the first n frames. These often 
            have a lot of false positives.
        overwrite : bool, default=True
             Overwrite existing CoordinateArrays or not.
        
        Returns
        -------
        CoordinateArrays  
        
        """
        from .coordinatearrays import create_coordarray
        
        if Path(filepath).suffix != '.darr':
            filepath = filepath + '.darr'
        Path(filepath).parent.mkdir(parents=True, exist_ok=True)
        
        coordsarray = create_coordarray(filepath,
                                        framewidth=self._framewidth,
                                        frameheight=self._frameheight,
                                        metadata=metadata,
                                        overwrite=overwrite)
        
        empty = np.zeros((0,2), dtype=np.uint16)
        coords = (c if i >= ignore_firstnframes else empty for i,c in
                  enumerate(self.find_nonzero()))
        coordsarray.iterappend(coords)
        
        return coordsarray

    def find_nonzero(self):
        """Yields the locations of non-zero pixels.

        If the frame is a color frame, non-zero means that a pixel
        does not have the value (0,0,0).

        Yields
        ------
        Iterator that generates shape (N, 2) arrays, where N is the number
            of non-zero pixels.

        """
        for frame in self._frames:
            if frame.ndim == 3:
                frame = (frame!=0).sum(axis=2, dtype=frame.dtype)
            idx = cv.findNonZero(frame)
            if idx is None:
                idx = np.zeros((0,2), dtype=np.uint16)
            else:
                idx = idx[:, 0, :]
            yield idx

    @frameiterator
    def morphologyex(self, morphtype='open', kernelsize=2, iterations=1):
        """Performs advanced morphological transformations on frames.

        Can perform advanced morphological transformations using an erosion
        and dilation as basic operations.

        In case of multi-channel images, each channel is processed
        independently.

        Parameters
        ----------
        morphtype : {'open', 'erode', 'dilate, 'close', 'gradient', 'tophat',
        'blackhat'}
            Type of transformation. Default is 'open', which is an erosion
            followed by a dilation.
        kernelsize : int, default=2
            Size of kernel in 1 dimension.
        iterations : int, default=1
            Number of times erosion and dilation are applied.

        Yields
        ------
        Frames
            Iterator that generates transformed image frames.

        """
        morphtypes={'erode': cv.MORPH_ERODE,
                    'dilate': cv.MORPH_DILATE,
                    'open': cv.MORPH_OPEN,
                    'close': cv.MORPH_CLOSE,
                    'gradient': cv.MORPH_GRADIENT,
                    'tophat': cv.MORPH_TOPHAT,
                    'blackhat': cv.MORPH_BLACKHAT}
        morphnum = morphtypes.get(morphtype, None)
        if morphnum is None:
            raise ValueError(f'`{morphtype}` is not a valid morphtype')
        kernel = np.ones((kernelsize, kernelsize), np.uint8)
        for frame in self._frames:
            yield cv.morphologyEx(frame, morphnum, kernel,
                                  iterations=iterations)

    @frameiterator
    def add_weighted(self, alpha, frames, beta, gamma=0):
        """Calculates the weighted sum of frames from self and the frames
        of the object specified by the `frames` parameter.

        Parameters
        ----------
        alpha : float
            Weight of the frames of self.
        frames : frame iterator
            The other source of input frames.
        beta : float
            Weight of the frames of the other frame iterator, specified by
            the `frames` parameter.
        gamma : float, optional
            Scalar added to each sum.

        Yields
        ------
        Frames
            Iterator that generates summed image frames.

        """
        for frame1, frame2 in zip(self._frames, frames):
            yield cv.addWeighted(src1=frame1, alpha=alpha, src2=frame2,
                                 beta=beta, gamma=gamma)

    @frameiterator
    def apply_backgroundsegmenter(self, bgs, fgmask=None, learningRate=-1.0,
                                  roi=None, nroi=None):
        """Compute foreground masks based on input sequence of frames.

        Parameters
        ----------
        bgs : Subclass of BaseBackgroundSubtractor
            Instance of one of Birdwatcher's BackgroundSubtractor classes,
            such as BackgroundSubtractorMOG2.
        fgmask : numpy array image, optional
            The output foreground mask as an 8-bit binary image.
        learningRate : float, optional
            The value between 0 and 1 that indicates how fast the background
            model is learnt. The default negative parameter value (-1.0) makes
            the algorithm to use some automatically chosen learning rate. 0
            means that the background model is not updated at all, 1 means
            that the background model is completely reinitialized from the
            last frame.
        roi : (int, int, int, int), optional
            Region of interest. Only look at this rectangular region. h1,
            h2, w1, w2.
        nroi : (int, int, int, int), optional
            Not region of interest. Exclude this rectangular region. h1,
            h2, w1, w2.

        Yields
        ------
        Frames
            Iterator that generates foreground masks.

        """
        if roi is not None:
            firstframe = self.peek_frame()
            completeframe = np.zeros((firstframe.shape[0],
                                      firstframe.shape[1]), dtype=np.uint8)
        for frame in self._frames:
            if roi is not None:
                h1,h2,w1,w2 = roi
                frame = frame[h1:h2, w1:w2]
            mask = bgs.apply(frame=frame, fgmask=fgmask,
                             learningRate=learningRate)
            if roi is not None:
                completeframe[h1:h2, w1:w2] = mask
                mask = completeframe
            if nroi is not None:
                h1,h2,w1,w2 = nroi
                mask[h1:h2, w1:w2] = 0
            yield mask

    @frameiterator
    def crop(self, h1, h2, w1, w2):
        """Crops frames to a smaller size.

        Parameters
        ----------
        h1 : int
            Top pixel rows.
        h2 : int
            Bottom pixel row.
        w1 : int
            Left pixel column.
        w2 : int
            Right pixel colum.

        Yields
        ------
        Frames
            Iterator that generates cropped frames.

        """
        for frame in self._frames:
            yield frame[h1:h2, w1:w2]

    @frameiterator
    def absdiff_frame(self, frame):
        """Subtract static image frame from frame iterator.

        Parameters
        ----------
        frame : ndarray frame
            Fixed image frame that will be subtracted from each frame of the
            frame iterator.

        Yields
        ------
        Frames
            Iterator that generates absolute difference frames.

        """
        for frame_self in self._frames:
            yield cv.absdiff(frame_self, frame)

    @frameiterator
    def threshold(self, thresh,	maxval=255,	threshtype='tozero'):
        """Thresholds frames at value `thresh`.

        Parameters
        ----------
        thresh : int
            Threshold value.
        maxval : int, default=255
        	Maximum value to use with the THRESH_BINARY and THRESH_BINARY_INV
            thresholding types.
        threshtype : {'tozero', 'tozero_inv', 'binary', 'binary_inv', 'trunc',
                     'mask', 'otsu', 'triangle'}, optional
            Thresholding type. The default is 'tozero', which means that
            everything below `thresh` will be set to zero. See doc OpenCV.

        Yields
        ------
        Frames
            Iterator that generates thresholded frames.

        """
        threshtypes = {
        'binary': cv.THRESH_BINARY,
        'binary_inv': cv.THRESH_BINARY_INV,
        'trunc': cv.THRESH_TRUNC,
        'tozero': cv.THRESH_TOZERO,
        'tozero_inv': cv.THRESH_TOZERO_INV,
        'mask': cv.THRESH_MASK,
        'otsu': cv.THRESH_OTSU,
        'triangle': cv.THRESH_TRIANGLE
        }
        threshtype = threshtypes[threshtype]
        for frame in self._frames:
            yield cv.threshold(src=frame, thresh=thresh,
                               maxval=maxval, type=threshtype)[1]

    @frameiterator
    def resize(self, dsize, interpolation='linear'):
        """Resizes frames.

        Parameters
        ----------
        dsize : tuple
            Destination size (width, height).
        interpolation : {'linear', 'nearest', 'area', 'cubic', 'lanczos4'},
                        optional
            Interpolation method:
            linear - a bilinear interpolation (used by default).
            nearest - a nearest-neighbor interpolation.
            area - resampling using pixel area relation. It may be a preferred
            method for image decimation, as it gives moire-free
            results. But when the image is zoomed, it is similar to
            the nearest method.
            cubic - a bicubic interpolation over 4x4 pixel neighborhood.
            lanczos4 - a Lanczos interpolation over 8x8 pixel neighborhood.

        Yields
        ------
        Frames
            Iterator that generates resized frames.

        """
        interptypes = {
            'nearest': cv.INTER_NEAREST,
            'linear': cv.INTER_LINEAR,
            'area': cv.INTER_AREA,
            'cubic': cv.INTER_CUBIC,
            'lanczos4': cv.INTER_LANCZOS4
        }
        interpolation = interptypes[interpolation]
        for frame in self._frames:
            yield cv.resize(src=frame, dsize=dsize, fx=0, fy=0,
                            interpolation=interpolation)

    @frameiterator
    def resizebyfactor(self, fx, fy, interpolation='linear'):
        """Resizes frames by a specified factor.

        Parameters
        ----------
        fx : float
            Scale factor along the horizontal axis.
        fy : float
            Scale factor along the vertical axis.
        interpolation : {'linear', 'nearest', 'area', 'cubic', 'lanczos4'},
                        optional
            Interpolation method:
            linear - a bilinear interpolation (used by default).
            nearest - a nearest-neighbor interpolation.
            area - resampling using pixel area relation. It may be a preferred
            method for image decimation, as it gives moire-free
            results. But when the image is zoomed, it is similar to
            the nearest method.
            cubic - a bicubic interpolation over 4x4 pixel neighborhood.
            lanczos4 - a Lanczos interpolation over 8x8 pixel neighborhood.

        Yields
        ------
        Frames
            Iterator that generates resized frames.

        """
        interptypes = {
            'nearest': cv.INTER_NEAREST,
            'linear': cv.INTER_LINEAR,
            'area': cv.INTER_AREA,
            'cubic': cv.INTER_CUBIC,
            'lanczos4': cv.INTER_LANCZOS4
        }
        interpolation = interptypes[interpolation]
        for frame in self._frames:
            yield cv.resize(src=frame, dsize=(0,0), fx=fx, fy=fy,
                            interpolation=interpolation)

    def find_contours(self, retrmode='tree', apprmethod='simple',
                      offset=(0, 0)):
        """Finds contours in frames.

        Contours can only be performed on gray frames. Use threshold or edge
        detection before applying contours for optimal results.

        Parameters
        ----------
        retrmode : str, optional
        apprmethod : str, optional
        offset : (int, int), optional

        Yields
        ------
        Generator
            Iterator that generates tuples (contours, hierarchy), with
            contours as a tuple of arrays, and hierarchy as an array denoting
            the parent-child relationship between contours.

        """
        retrmode = {'tree': cv.RETR_TREE,
                    'external': cv.RETR_EXTERNAL,
                    'list': cv.RETR_LIST,
                    'ccomp': cv.RETR_CCOMP,
                    'floodfill': cv.RETR_FLOODFILL}[retrmode]
        apprmethod = {'none': cv.CHAIN_APPROX_NONE,
                      'simple': cv.CHAIN_APPROX_SIMPLE,
                      'tc89_l1': cv.CHAIN_APPROX_TC89_L1,
                      'tc89_kcos': cv.CHAIN_APPROX_TC89_KCOS}[apprmethod]
        for frame in self._frames:
            yield cv.findContours(frame, mode=retrmode, method=apprmethod, 
                                  offset=offset)

    def calc_meanframe(self, dtype=None):
        if self.nchannels == 1:
            meanframe = framegray(self.frameheight, self.framewidth, value=0,
                                  dtype='float64')
        else:
            meanframe = framecolor(self.frameheight, self.framewidth, color=
                                   (0, 0, 0), dtype='float64')
        for i, frame in enumerate(self._frames):
            meanframe += frame
        meanframe /= (i+1)
        if dtype is None:
            dtype = self._dtype
        return meanframe.astype(dtype)

    # TODO Frames should have a frame rate
    # TODO Exception handling
    def show(self, framerate=25):
        """Shows frames in a video window.

        Iterates through Frames and displaying each frame in a seperate
        window. Press 'q' to quit the video before the end.

        Parameters
        ----------
        framerate : int, default=25

        Notes
        -----
        Frames iterator is (partly) empty after using 'show'.

        """
        waitkey = int(round(1000 / framerate))
        for frame in self._frames:
            cv.imshow('frame', frame)
            if cv.waitKey(waitkey) & 0xFF == ord('q'):
                break
        cv.destroyAllWindows()


class FramesColor(Frames):
    """An iterator that yields color frames.

    This class inherits from Frames, and hence has all its methods.

    """

    def __init__(self, nframes, height, width, color=(0, 0, 0), 
                 dtype='uint8'):
        """Creates an iterator that yields color frames.

        Parameters
        ----------
        nframes : int
            Number of frames to be produced.
        height : int
            Height of frame.
        width : int
            Width of frame.
        color : tuple of ints, optional
            Fill value of frame (r, g, b). The default (0, 0, 0) color is
            black.
        dtype : numpy dtype, default='uint8'
            Dtype of frame.

        Returns
        -------
        Iterator of numpy ndarrays

        """
        frame = framecolor(height=height, width=width, color=color,
                           dtype=dtype)
        frames = (frame.copy() for _ in range(nframes))
        super().__init__(frames=frames)


class FramesGray(Frames):
    """An iterator that yields gray frames.

    This class inherits from Frames, and hence has all its methods.

    """

    def __init__(self, nframes, height, width, value=0, dtype='uint8'):
        """Creates an iterator that yields gray frames.

        Parameters
        ----------
        nframes : int
            Number of frames to be produced.
        height : int
            Height of frame.
        width : int
            Width of frame.
        value : int, optional
            Fill value of frame. The default (0) is black.
        dtype : numpy dtype, default='uint8'
            Dtype of frame.

        Returns
        -------
        Iterator of numpy ndarrays

        """

        frame = framegray(height=height, width=width, value=value,
                          dtype=dtype)
        frames = (frame.copy() for _ in range(nframes))
        super().__init__(frames=frames)


class FramesNoise(Frames):
    """An iterator that yields noise frames.

    This class inherits from Frames, and hence has all its methods.

    """

    def __init__(self, nframes, height, width, dtype='uint8'):
        """Creates an iterator that yields gray frames.

        Parameters
        ----------
        nframes : int
            Number of frames to be produced.
        height : int
            Height of frame.
        width : int
            Width of frame.
        dtype : numpy dtype, default='uint8'
            Dtype of frame.

        Returns
        -------
        Iterator of numpy ndarrays

        """

        frames = (framenoise(height=height, width=width, dtype=dtype)
                  for _ in range(nframes))
        super().__init__(frames=frames)


def framegray(height, width, value=0, dtype='uint8'):
    """Creates a gray frame.

    Parameters
    ----------
    height : int
        Height of frame.
    width : int
        Width of frame.
    value : int, optional
        Fill value of frame. The default (0) is black.
    dtype : numpy dtype, default='uint8'
        Dtype of frame.

    Returns
    -------
    numpy ndarray

    """
    return np.ones((height, width), dtype=dtype) * value


def framecolor(height, width, color=(0, 0, 0), dtype='uint8'):
    """Creates a color frame.

    Parameters
    ----------
    height : int
        Height of frame.
    width : int
        Width of frame.
    color : tuple of ints, optional
        Fill value of frame (r, g, b). The default (0, 0, 0) color is black.
    dtype : numpy dtype, default='uint8'
        Dtype of frame.

    Returns
    -------
    numpy ndarray

    """
    return np.ones((height, width, 3), dtype=dtype) * np.asanyarray(color, dtype=dtype)


def framenoise(height, width, dtype='uint8'):
    """Creates a noise frame.

    Parameters
    ----------
    height : int
        Height of frame.
    width : int
        Width of frame.
    dtype : numpy dtype, default='uint8'
        Dtype of frame.

    Returns
    -------
    numpy ndarray

    """
    return np.random.randint(0, 255, (height, width, 3), dtype=dtype)


def create_frameswithmovingcircle(nframes, width, height,
                                  framecolor=(0, 0, 0),
                                  circlecolor=(255, 100, 0), radius=6,
                                  thickness=2, linetype=8, dtype='uint8'):

    frames = FramesColor(nframes=nframes,  width=width, height=height,
                         color=framecolor, dtype=dtype)
    centers = zip(np.linspace(0, width, nframes),
                  np.linspace(0, height, nframes))
    return frames.draw_circles(centers, color=circlecolor, radius=radius,
                               thickness=thickness, linetype=linetype)



import os
import unittest
import tempfile
import shutil
from pathlib import Path

import numpy as np
from numpy.testing import assert_array_equal

import birdwatcher as bw
from birdwatcher.coordinatearrays import _archive


class TestCoordinateArrays(unittest.TestCase):

    def setUp(self):
        self.tempdirname1 = tempfile.mkdtemp()
        fh, self.tempvideoname = tempfile.mkstemp()
        os.close(fh)
        metadata = {'avgframerate': 5}
        self.ca1 = bw.create_coordarray(path=self.tempdirname1,
                                    framewidth=1080, frameheight=720,
                                    metadata=metadata, overwrite=True)
        self.ca1.iterappend([((1,2),(3,4)),((5,6),(7,8))])

    def tearDown(self):
        shutil.rmtree(self.tempdirname1)
        if Path(self.tempvideoname).exists():
            Path(self.tempvideoname).unlink()

    def test_index(self):
        assert_array_equal(self.ca1[1], np.array([[5,6],[7,8]]))

    def test_get_frame(self):
        frame = self.ca1.get_frame(1)
        self.assertEqual(np.sum(frame), 2)
        self.assertEqual(frame[6,5], 1)
        self.assertEqual(frame[8,7], 1)
        self.assertTupleEqual(frame.shape, (720,1080))

    def test_iter_frames(self):
        for frame in self.ca1.iter_frames():
            self.assertEqual(np.sum(frame), 2)
            self.assertTupleEqual(frame.shape, (720, 1080))

    def test_tovideo(self):
        self.ca1.tovideo(self.tempvideoname, framerate=5)

    def test_inferframerate(self):
        self.ca1.tovideo(self.tempvideoname)

    def test_coordcount(self):
        cc = self.ca1.get_coordcount()
        assert_array_equal(cc, np.array([2,2]))

    def test_coordmean(self):
        cm = self.ca1.get_coordmean()
        assert_array_equal(cm, np.array([[2,3],[6,7]]))

    def test_coordmedian(self):
        self.ca1.iterappend([((1,2),(3,4),(8,9))])
        cmd = self.ca1.get_coordmedian()
        assert_array_equal(cmd, np.array([[2,3],[6,7],[3,4]]))


class TestArchivedCoordinateArrays(unittest.TestCase):

    def setUp(self):
        self.tempdirname1 = Path(tempfile.mkdtemp())
        self.archivename = self.tempdirname1.parent / (self.tempdirname1.name + '.tar.xz')
        metadata = {'avgframerate': 5}
        self.ca1 = bw.create_coordarray(path=self.tempdirname1,
                                        framewidth=1080, frameheight=720,
                                        metadata=metadata, overwrite=True)
        self.ca1.iterappend([((1,2),(3,4)),((5,6),(7,8))])
        
    def tearDown(self):
        if self.tempdirname1.exists():
            shutil.rmtree(self.tempdirname1)
        if self.archivename.exists():
            self.archivename.unlink()
    
    def test_archive(self):
        _archive(self.ca1)
        self.assertTrue(self.archivename.exists())
        self.assertFalse(self.tempdirname1.exists())
    
    def test_open_archived(self):
        coords1 = self.ca1[1]
        _archive(self.ca1)
        with bw.open_archivedcoordinatedata(self.archivename) as ca2:
            assert_array_equal(ca2[1], coords1)

    def test_extract_archived(self):
        coords1 = self.ca1[1]
        _archive(self.ca1)
        ca2 = bw.extract_archivedcoordinatedata(self.archivename)
        assert_array_equal(ca2[1], coords1)


class TestMoveCoordinateArrays(unittest.TestCase):

    def setUp(self):
        self.tempdirname1 = tempfile.mkdtemp()
        self.tempdirname2 = tempfile.mkdtemp()
        path = Path(self.tempdirname1)/'even.darr'
        self.ca1 = bw.create_coordarray(path=path, framewidth=1080,
                                     frameheight=720, overwrite=True)
        self.ca1.iterappend([((1, 2), (3, 4)), ((5, 6), (7, 8))])

    def tearDown(self):
        shutil.rmtree(self.tempdirname1)
        shutil.rmtree(self.tempdirname2)

    def test_movecoordinatearrays(self):
        bw.move_coordinatearrays(self.tempdirname1, self.tempdirname2)
        ca2 = bw.CoordinateArrays(Path(self.tempdirname2) / 'even.darr')









"""This module contains objects and functions helpfull for determining which settings result in optimal movement detection.

"""

import json
from pathlib import Path

import numpy as np
import pandas as pd
import seaborn as sns
import matplotlib.pyplot as plt

import birdwatcher as bw
import birdwatcher.movementdetection as md
from birdwatcher.utils import product_dict


__all__ = ['ParameterSelection', 'apply_all_parameters', 
           'load_parameterselection', 'product_dict']


class ParameterSelection():
    """A Pandas dataframe with movement detection results of various parameter 
    settings associated with a (fragment of a) Videofilestream.

    """
    # colors in BGR
    colors = [('orange', [0, 100, 255]),
              ('blue', [255, 0, 0]),
              ('red', [0, 0, 255]),
              ('lime', [0, 255, 0]),
              ('cyan', [255, 255, 0]),
              ('magenta', [255, 0, 255])]

    def __init__(self, df, videofilepath, bgs_type, 
                 startat, nframes, roi, nroi, path=None):
        self.df = df
        self.vfs = bw.VideoFileStream(videofilepath)
        self.bgs_type = bgs_type
        self.startat = startat
        self.nframes = nframes
        self.roi = roi
        self.nroi = nroi
        self.path = path

    def get_info(self):
        return {'vfs': str(self.vfs.filepath),
                'bgs_type': self.bgs_type,
                'startat': self.startat,
                'nframes': self.nframes,
                'roi': self.roi,
                'nroi': self.nroi}

    def get_videofragment(self):
        """Returns video fragment as Frames.

        NOTE: the whole frames are returned. If a region of interest (roi or 
        nroi) is specified, this is not visible in the videofragment.

        """
        return self.vfs.iter_frames(startat=self.startat, 
                                    nframes=self.nframes)

    def get_parameters(self, selection='multi_only'):
        """Returns the parameter settings used for movement detection.

        Parameters
        ----------
        selection : {'all', multi_only'}
            Specify which selection of parameters is returned:
            all : returns all parameters and their settings.
            multi_only : returns only parameters for which multiple values 
            have been used to run movement detection.

        Returns
        ------
        dict
            With parameters as keys, and each value contains a list of the 
            settings used for movement detection.

        """
        paramkeys = (set(self.df.columns) - 
                     set(['framenumber', 'pixel', 'coords']))
        all_parameters = {k:self.df[k].unique().tolist() for k in paramkeys}

        if selection == 'all':
            return all_parameters
        elif selection == 'multi_only':
            return {k:all_parameters[k] for k in paramkeys if 
                    len(all_parameters[k])>1}
        else:
            raise Exception(f"'{selection}' is not recognized. Please "
                            "choose between 'all' and 'multi_only'.")

    def plot_parameters(self, rows, cols, default_values):
        """Returns a figure from seaborn with subplots.

        Usefull to look at different location detection results from two 
        parameters with various values.

        Parameters
        ----------
        rows : str
            One parameter tested with multiple values.
        cols : str
            A second parameter tested with multiple values.
        default_values : dict
            All parameters that are tested with multiple settings, should be 
            added to a dictionary with each parameter as key, and the default 
            as value.

        Returns
        ------
        FacedGrid
            A seaborn object managing multiple subplots.

        """
        self._check_multi_only(default_values)
        other_values = {key:default_values[key] for key in 
                        default_values.keys()-[rows, cols]}
        df_selection = self._select_data(**other_values)

        # plot with seaborn
        g = sns.relplot(x="framenumber", y="pixel", hue="coords", style=None, 
                        col=cols, row=rows, kind="line", data=df_selection, 
                        height=3, aspect=2)
        g.figure.suptitle(str(other_values), fontsize=15, x=0.51, y=1.05)

        return g

    def batch_plot_parameters(self, default_values, overwrite=False):
        """Saves multiple figures with subplots of all combinations of 
        parameters.

        The figures are saved in a folder 'figures' in the same directory as 
        the associated ParameterSelection file. Multiple rounds of plotting 
        with different default values can be easily saved in new folders with 
        a number added to the foldername.

        Parameters
        ----------
        default_values : dict
            All parameters that are tested with multiple settings, should be 
            added to a dictionary with each parameter as key, and the default 
            as value.
        overwrite : bool, default=False
            If False, an integer number (1,2,3,etc.) will be added as suffix 
            to the foldername, if the filepath already exists.

        """
        path = self._create_path(self.path, 'figures', overwrite)

        # save default values as txt file
        with open(path / 'default_values.txt', 'w') as f:
            for key, val in default_values.items():
                f.write(f"{key}: {val}\n")

        # plot and save each combination of two parameters
        settings = self.get_parameters('multi_only')
        for rows, _ in settings.items():
            for cols, _ in settings.items():
                if rows != cols:
                    filename = f'{self.vfs.filepath.stem}_{rows}_{cols}.png'
                    g = self.plot_parameters(rows, cols, default_values)
                    g.savefig(path / filename)
                    plt.close(g.figure)
        print(f"The figures are saved in {path}")

    def draw_multiple_circles(self, all_settings, radius=60, thickness=2):
        """Returns a Frames object with circles on the videofragment.

        It is possible to plot multiple circles on the videofragment to see 
        the results from different parameter settings.

        Parameters
        ----------
        settings : dict
            All parameters that are tested with multiple settings, should be 
            added to a dictionary with each parameter as key, and the value(s) 
            that should be superimposed on the video added as list.
        radius : int, default=60
            Radius of circle.
        thickness : int, default=2
            Line thickness.

        Returns
        -------
        Frames, DataFrame
            Iterator that generates frames with multiple circles, and a Pandas 
            DataFrame with the settings for each color of the circles.

        """
        self._check_stats_is_not_count()
        self._check_multi_only(all_settings)
        self._check_number_of_colorcombinations(all_settings)

        frames = self.get_videofragment().draw_framenumbers()
        colorspecs = {}
        for i, settings in enumerate(product_dict(**all_settings)):
            colorspecs[self.colors[i][0]] = settings
            df_selection = self._select_data(**settings)
            iterdata = (df_selection.set_index(['framenumber', 'coords'])
                        .loc[:, 'pixel'].unstack().values)
            frames = frames.draw_circles(iterdata, radius=radius,
                                         color=self.colors[i][1],
                                         thickness=thickness)

        return frames, pd.DataFrame(colorspecs)

    def save_parameters(self, path, foldername=None, overwrite=False):
        """Save results of all parameter settings as .csv file.

        Often several rounds of parameter selection per videofragment will be 
        done with different parameter settings. For this, the same foldername 
        could be used, in which case a number is added automatically as suffix 
        to display the round.

        Parameters
        ----------
        path : str
            Path to disk-based directory that should be written to.
        foldername : str, optional
            Name of the folder the data should be written to.
        overwrite : bool, default=False
            If False, an integer number (1,2,3,etc.) will be added as suffix 
            to the foldername, if the filepath already exists.

        """
        if foldername is None:
            foldername = f'params_{self.vfs.filepath.stem}'
        path = self._create_path(path, foldername, overwrite)
        self.path = str(path)

        # add extra information to saved .csv file
        json_info = json.dumps(self.get_info())
        self.df.to_csv(path / 'parameterselection.csv', index_label=json_info)

        # save parameter settings as readme file
        with open(path / 'readme.txt', 'w') as f:
            for key, val in self.get_parameters('all').items():
                f.write(f"{key}: {val}\n")


    def _create_path(self, path, foldername, overwrite):
        """Useful for creating a path with a number added as suffix in case 
        the folder already exists.

        Parameters
        ----------
        path : str
            Path to disk-based directory that should be written to.
        foldername : str, optional
            Name of the folder the data should be written to.
        overwrite : bool, default=False
            If False, an integer number (1,2,3,etc.) will be added as suffix 
            to the foldername, if the filepath already exists. 

        """
        path = Path(path) / foldername

        if not overwrite:
            i = 1
            while path.exists():
                i += 1
                path = path.parent / f'{foldername}_{i}'
        Path(path).mkdir(parents=True, exist_ok=overwrite)

        return path

    def _select_data(self, **kwargs):
        """Returns a copy with a selection of the dataframe.

        **kwargs can be a dictionary, with keys matching column names. The
        value of each key will be selected from a copy of the dataframe.

        """
        df = self.df.copy()
        for key, value in kwargs.items():
            df = df.loc[df[key]==value]

        return df

    def _check_multi_only(self, inputdict):
        """Check if all parameters that are tested with multiple settings, are 
        included in the inputdic.

        """
        multi_only = self.get_parameters('multi_only')
        if multi_only.keys() != inputdict.keys():
            raise Exception("Make sure the input dictionary contains all keys"
                            f" with multiple values: {multi_only.keys()}")

    def _check_number_of_colorcombinations(self, settings):
        """Check if the number of settings combinations does not exceed the
        number of possible colors.

        """
        n_combinations = len(list(product_dict(**settings)))
        n_colors = len(self.colors)
        if n_combinations > n_colors:
            raise Exception(
                f"The number of settings combinations is {n_combinations}, "
                f"but a maximum of {n_colors} circles can be plotted. Reduce "
                "the number of settings combinations, or add new colors to "
                "the class attribute 'colors' to be able to plot more "
                "circles.")
            
    def _check_stats_is_not_count(self):
        if self.df.coords.unique()[0] == 'count':
            raise Exception("This function cannot be executed with count "
                            "statistics.")


def apply_all_parameters(vfs, all_settings, startat=None, nframes=None, 
                         roi=None, nroi=None, use_stats='mean',
                         bgs_type=bw.BackgroundSubtractorMOG2, 
                         reportprogress=50):
    """Run movement detection with each set of parameters.

    Parameters
    ----------
    vfs : VideoFileStream
        A Birdwatcher VideoFileStream object.
    all_settings : {dict, dict}
        Dictionary with two dictionaries. One 'bgs_params' with the parameter 
        settings from the BackgroundSubtractor and another 'processing' 
        dictionary with settings for applying color, resizebyfactor, blur and 
        morphologyex manipulations.
    startat : str, optional
        If specified, start at this time point in the video file. You can use 
        two different time unit formats: sexagesimal 
        (HOURS:MM:SS.MILLISECONDS, as in 01:23:45.678), or in seconds.
    nframes  : int, optional
        Read a specified number of frames.
    roi : (int, int, int, int), optional
        Region of interest. Only look at this rectangular region. h1,
        h2, w1, w2.
    nroi : (int, int, int, int), optional
        Not region of interest. Exclude this rectangular region. h1,
        h2, w1, w2.
    use_stats : str, default='mean'
        As default the mean pixel per frame is calculated. Another option is 
        to use 'count' to calculate to number of pixels per frame.    
    bgs_type: BackgroundSubtractor
        This can be any of the BackgroundSubtractors in Birdwatcher, e.g. 
        BackgroundSubtractorMOG2, BackgroundSubtractorKNN, 
        BackgroundSubtractorLSBP.
    reportprogress: int or bool, default=50
        The input integer represents how often the progress of applying each 
        combination of settings is printed. Use False, to turn off 
        reportprogress.

    """
    if reportprogress:
        import datetime
        starttime = datetime.datetime.now()
        n = 0

    list_with_dfs = []
    all_combinations = product_dict(**all_settings['bgs_params'],
                                    **all_settings['processing'])

    for settings in all_combinations:
        frames = md.apply_settings(vfs, settings, startat, nframes, roi, nroi, 
                                   bgs_type)
        coordinates = frames.find_nonzero()
    
        # calculate mean or count statistics of coordinates
        if use_stats == 'mean':
            coordstats = np.array([c.mean(0) if c.size>0 else (np.nan, np.nan) 
                                   for c in coordinates])
            settings['coords'] = ['x', 'y']
        elif use_stats == 'count':
            coordstats = np.array([c.shape[0] for c in coordinates])
            settings['coords'] = ['count']
        else:
            raise Exception("Make sure use_stats has 'mean' or 'count' as "
                            "input value")
            
        # add results as pandas DataFrame
        columns = pd.MultiIndex.from_frame(pd.DataFrame(settings))
        df = pd.DataFrame(coordstats, columns=columns)
        list_with_dfs.append(df)

        if reportprogress:
            n += 1
            if n % reportprogress == 0:
                diff = datetime.datetime.now() - starttime 
                print(f'{n} combinations of settings applied in'
                      f' {str(diff).split(".")[0]} hours:min:sec')

    # create long-format DataFrame
    df = pd.concat(list_with_dfs, axis=1)
    df.index.name = 'framenumber'
    df = (df.stack(df.columns.names, future_stack=True)
          .reset_index()  # stack all column levels
          .rename({0: 'pixel'}, axis=1))

    return ParameterSelection(df, vfs.filepath, str(bgs_type), startat, 
                              nframes, roi, nroi)

def load_parameterselection(path):
    """Load a parameterselection.csv file.

    Parameters
    ----------
    path : str
        Name of the directory where parameterselection.csv is saved.

    Returns
    -------
    ParameterSelection

    """
    filepath = Path(path) / 'parameterselection.csv' 
    df = pd.read_csv(filepath, index_col=0, engine='python')
    info = json.loads(df.index.names[0])
    df.index.name = None

    return ParameterSelection(df, info['vfs'], info['bgs_type'], 
                              info['startat'], info['nframes'], 
                              info['roi'], info['nroi'], path)


from .video import *
from .backgroundsubtraction import *
from .coordinatearrays import *
from .frames import *

from ._version import get_versions
__version__ = get_versions()['version']
del get_versions

from .tests import test


import unittest
import birdwatcher as bw
from birdwatcher.frames import create_frameswithmovingcircle


class TestBackgroundSubtractorKNN(unittest.TestCase):

    def test_KNNdefaultinstantiation(self):
        bw.BackgroundSubtractorKNN()

    def test_KNNparams(self):
        bgs = bw.BackgroundSubtractorKNN(History=10)
        self.assertEqual(bgs.get_params()['History'], 10)

    def test_KNNapply(self):
        bgs = bw.BackgroundSubtractorKNN(History=10)
        frames = create_frameswithmovingcircle(nframes=5, width=1080,
                                               height=720)
        for fg in frames.apply_backgroundsegmenter(bgs,  
                                                   roi=(10, 710, 10, 500),
                                                   nroi=(20,30,20,30)):
            pass


class TestBackgroundSubtractorMOG2(unittest.TestCase):

    def test_MOG2defaultinstantiation(self):
        bw.BackgroundSubtractorMOG2()

    def test_MOG2params(self):
        bgs = bw.BackgroundSubtractorMOG2(History=10)
        self.assertEqual(bgs.get_params()['History'], 10)

    def test_MOG2apply(self):
        bgs = bw.BackgroundSubtractorMOG2(History=10)
        frames = create_frameswithmovingcircle(nframes=5, width=1080,
                                               height=720)
        for fg in frames.apply_backgroundsegmenter(bgs,  
                                                   roi=(10, 710, 10, 500),
                                                   nroi=(20, 30, 20, 30)):
            pass


class TestBackgroundSubtractorLSBP(unittest.TestCase):

    def test_LSBPdefaultinstantiation(self):
        bw.BackgroundSubtractorLSBP()

    def test_LSBPparams(self):
        bgs = bw.BackgroundSubtractorLSBP(nSamples=10)
        self.assertEqual(bgs.get_params()['nSamples'], 10)

    def test_LSBPapply(self):
        bgs = bw.BackgroundSubtractorLSBP()
        frames = create_frameswithmovingcircle(nframes=5, width=1080,
                                               height=720)
        for fg in frames.apply_backgroundsegmenter(bgs):
            pass


from unittest import TestLoader, TextTestRunner, TestSuite

from . import test_videoinput
from . import test_frameprocessing
from . import test_backgroundsubtraction
from . import test_coordinatearrays
from . import test_plotting
from . import test_movementdetection
from . import test_parameterselection


modules = [test_videoinput, test_frameprocessing, test_backgroundsubtraction, 
           test_coordinatearrays, test_plotting, test_movementdetection, 
           test_parameterselection]


def test(verbosity=1):
    suite = TestSuite()
    for module in modules:
        suite.addTests(TestLoader().loadTestsFromModule(module))
    return TextTestRunner(verbosity=verbosity).run(suite)


import itertools
import os, sys
import pathlib
import shutil
import tempfile
import time
from contextlib import contextmanager


def datetimestring():
    return time.strftime('%Y%m%d%H%M%S')


def product_dict(**kwargs):
    """Generates a Cartesian product of dictionary values.
    
    """
    keys = kwargs.keys()
    vals = kwargs.values()
    for instance in itertools.product(*vals):
        yield dict(zip(keys, instance))


@contextmanager
def tempdir(dirname='.', keep=False, report=False):
    """Yields a temporary directory which is removed when context is
    closed."""
    try:
        tempdirname = tempfile.mkdtemp(dir=dirname)
        if report:
            print('created tempdir {}'.format(tempdirname))
        yield pathlib.Path(tempdirname)
    except:
        raise
    finally:
        if not keep:
            shutil.rmtree(tempdirname)
            if report:
                print('removed temp dir {}'.format(tempdirname))


def peek_iterable(iterable):
    gen = (i for i in iterable)
    first = next(gen)
    return first, itertools.chain([first], gen)


def derive_filepath(filepath, append_string='', suffix=None, path=None):
    """Generate a file path based on the name and potentially path of the
    input file path.

    Parameters
    ----------
    filepath : str of pathlib.Path
        Path to file.
    append_string : str, optional
        String to append to file name stem.
    suffix : str, optional
        File extension to use. If None, the same as video file.
    path : str or pathlib.Path, optional
        Path to use. If None, use same path as video file.

    Returns
    -------
    pathlib.Path
        Path derived from video file path.

    """
    stem = filepath.stem
    if suffix is None:
        suffix = filepath.suffix
    filename = f'{stem}_{append_string}{suffix}'
    if path is None:
        dpath = filepath.parent / filename
    else:
        dpath = pathlib.Path(path) / filename
    return dpath


def progress(count, total, status=''):

    # The MIT License (MIT)
    # Copyright (c) 2016 Vladimir Ignatev
    #
    # Permission is hereby granted, free of charge, to any person obtaining
    # a copy of this software and associated documentation files (the "Software"),
    # to deal in the Software without restriction, including without limitation
    # the rights to use, copy, modify, merge, publish, distribute, sublicense,
    # and/or sell copies of the Software, and to permit persons to whom the Software
    # is furnished to do so, subject to the following conditions:
    #
    # The above copyright notice and this permission notice shall be included
    # in all copies or substantial portions of the Software.
    #
    # THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,
    # INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
    # PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE
    # FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT
    # OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE
    # OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

    bar_len = 60
    filled_len = int(round(bar_len * count / float(total)))

    percents = round(100.0 * count / float(total), 1)
    bar = '=' * filled_len + '-' * (bar_len - filled_len)

    sys.stdout.write('[%s] %s%s ...%s\r' % (bar, percents, '%', status))
    sys.stdout.flush() # As suggested by Rom Ruben (see: http://stackoverflow.com/questions/3173320/text-progress-bar-in-the-console/27871113#comment50529068_27871113)


def print_dirstructure(dirpath):
    """Prints the hierarchical structure of directories, starting at `dirpath`

    Parameters
    ----------
    dirpath : str or Path
        The top-level directory to start at.

    """
    for root, dirs, files in os.walk(dirpath):
        level = root.replace(dirpath, '').count(os.sep)
        indent = ' ' * 4 * (level)
        print(f'{indent}{os.path.basename(root)}/')
        subindent = ' ' * 4 * (level + 1)
        for f in files:
            if pathlib.Path(f).is_dir():
                print(f'{subindent}{f}')


def walk_paths(dirpath, extension='.*'):
    """Walks recursively over contents of `dirpath` and yield contents as
    pathlib Path objects, potentially based on their `extension`.

    Parameters
    ----------
    dirpath : str or Path
        The top-level directory to start at.
    extension : str, optional
        Filter on this extension. The default includes all extensions.

    """
    dirpath = pathlib.Path(dirpath)
    if extension.startswith('.'):
        extension = extension[1:]
    for file in dirpath.rglob(f'*.{extension}'):
        yield file


import unittest
import birdwatcher as bw
import tempfile
import shutil

from pathlib import Path


class TestVideos(unittest.TestCase):

    def test_videoiterframes(self):
        vf = bw.testvideosmall()
        shape = (vf.frameheight, vf.framewidth, 3)
        for i,frame in enumerate(vf.iter_frames(), 1):
            self.assertEqual(frame.shape, shape)
        self.assertEqual(i, 497)

    def test_size(self):
        vf = bw.testvideosmall()
        self.assertEqual(vf.framewidth, 1280)
        self.assertEqual(vf.frameheight, 720)
        self.assertSequenceEqual(vf.framesize, (1280, 720))

    def test_countframes(self):
        vf = bw.testvideosmall()
        self.assertEqual(vf.count_frames(), 497)

    def test_nframes(self):
        vf = bw.testvideosmall()
        self.assertEqual(vf.nframes, 497)

    def test_duration(self):
        vf = bw.testvideosmall()
        self.assertEqual(vf.duration, float(19.880000))

    def test_getframe(self):
        vf = bw.testvideosmall()
        frame = vf.get_frame(100)
        self.assertSequenceEqual(frame.shape, (720,1280,3))
    
    def test_getframeat(self):
        vf = bw.testvideosmall()
        frame = vf.get_frameat('00:10.')
        self.assertSequenceEqual(frame.shape, (720,1280,3))

    def test_extractaudio(self):
        d = Path(tempfile.mkdtemp())
        p = d / 'even.wav'
        bw.testvideosmall().extract_audio(outputpath=p)
        shutil.rmtree(d)









"""This module contains classes and functions to work with video files. It
depends on FFmpeg.

"""


from pathlib import Path

import numpy as np
import cv2 as cv

from .ffmpeg import videofileinfo, iterread_videofile, count_frames, \
    get_frame, get_frameat, extract_audio
from .frames import frameiterator
from .utils import progress


__all__ = ['VideoFileStream', 'testvideosmall']


class VideoFileStream():
    """Video stream from file.

    This class can read video frames from a file.

    Parameters
    ----------
    filepath : str of pathlib.Path
        Path to videofile.
    streamnumber : int, optional
        Video stream number to use as input. Often there is just
        one video stream present in a video file (default=0), but
        if there are more you can use this parameter to specify
        which one you want.

    Examples
    --------
    >>> import birdwatcher as bw
    >>> vfs = bw.VideoFileStream('zebrafinchrecording.mp4')
    >>> frames = vfs.iter_frames() # create frame iterator
    >>> frames.togray().tovideo('zebrafinchrecording_gray.mp4')

    """

    def __init__(self, filepath, streamnumber=0):

        self.filepath = fp = Path(filepath)
        self.streamnumber = streamnumber
        if not fp.exists():
            raise FileNotFoundError(f'"{filepath}" does not exist')
        metadata = videofileinfo(fp)
        self._formatmetadata = metadata['format']
        self._streammetadata = metadata['streams'][streamnumber]


    def __iter__(self):
        return self.iter_frames()

    def get_info(self):
        """Provides a dictionary will all kinds of video info.

        Much of it is provided by ffprobe.

        Returns
        -------
            Dictionary with info.
        """
        return {'classname': self.__class__.__name__,
                'classarguments': {'filepath': str(self.filepath),
                                  'streamnumber': self.streamnumber},
                'framewidth': self.framewidth,
                'frameheight': self.frameheight,
                'formatmetadata': self.formatmetadata,
                'streammetadata': self.streammetadata}

    @property
    def avgframerate(self):
        """Average frame rate of video stream, as reported in the metadata
        of the video file."""
        ar = self.streammetadata['avg_frame_rate']
        return np.divide(*map(int, ar.split('/')))

    @property
    def duration(self):
        """Duration of video stream in seconds, as reported in the metadata
        of the video file."""
        return float(self.streammetadata['duration'])

    @property
    def formatmetadata(self):
        """Metadata of video file format as provided by ffprobe."""
        return self._formatmetadata

    @property
    def streammetadata(self):
        """Metadata of video stream as provided by ffprobe."""
        return self._streammetadata

    @property
    def framewidth(self):
        """Width in pixels of frames in video stream."""
        return self._streammetadata['width']

    @property
    def frameheight(self):
        """height in pixels of frames in video stream."""
        return self._streammetadata['height']

    @property
    def framesize(self):
        """tuple (frame width, frame height) in pixels in video stream."""
        return (self.framewidth, self.frameheight)

    @property
    def nframes(self):
        """Number of frames in video stream as reported in the metadata
        of the video file. Note that this may not be accurate. Use
        `count_frames` to measure the actual number (may take a lot of
        time)."""
        return int(self.streammetadata['nb_frames'])

    def count_frames(self, threads=8, ffprobepath='ffprobe'):
        """Count the number of frames in video file stream.

        This can be necessary as the number of frames reported in the
        video file metadata may not be accurate. This method requires
        decoding the whole video stream and may take a lot of time. Use the
        `nframes` property if you trust the video file metadata and want
        fast results.

        Parameters
        ----------
        threads : int, default=8
            The number of threads you want to devote to decoding.
        ffprobepath : str or Path, default='ffprobe'

        Returns
        -------
        int
            The number of frames in video file.

        """
        return count_frames(self.filepath, threads=threads,
                            ffprobepath=ffprobepath)

    def extract_audio(self, outputpath=None, overwrite=False, 
                      codec='pcm_s24le', channel=None, ffmpegpath='ffmpeg', 
                      loglevel='quiet'):
        """Extract audio as wav file.

        Parameters
        ----------
        outputpath : str or pathlib.Path, optional
            Filename and path to write audio to. The default is None, which
            means the same directory and name as the video file is used, but 
            then with '.wav' extension.
        overwrite : bool, default=False
            Overwrite if audio file exists or not.
        codec : str, default='pcm_s24le'
            ffmpeg audio codec, with 24-bit pcm as default output.
        channel : int, default=None
            Channel number to extract. The default None will extract all 
            channels.
        ffmpegpath : str or pathlib.Path, optional
            Path to ffmpeg executable. Default is `ffmpeg`, which means it
            should be in the system path.
        loglevel : {'quiet', 'panic', 'fatal', 'error', 'warning', 'info', 
                    'verbose', 'debug' ,'trace'}, optional
            Level of info that ffmpeg should print to terminal. Default is 
            'quiet'.

        """
        filepath = self.filepath
        return extract_audio(filepath=filepath, outputpath=outputpath,
                             overwrite=overwrite, codec=codec, 
                             channel=channel, ffmpegpath=ffmpegpath, 
                             loglevel=loglevel)

    @frameiterator
    def iter_frames(self, startat=None, nframes=None, color=True,
                    ffmpegpath='ffmpeg', reportprogress=False):
        """Iterate over frames in video.

        Parameters
        ----------
        startat : str, optional
            If specified, start at this time point in the video file. You
            can use two different time unit formats: sexagesimal
            (HOURS:MM:SS.MILLISECONDS, as in 01:23:45.678), or in seconds.
        nframes  : int, optional
            Read a specified number of frames.
        color : bool, default=True
            Read as a color frame (3 dimensional) or as a gray frame (2
            dimensional). Color conversions occur through ffmpeg.
        ffmpegpath : str or pathlib.Path, optional
            Path to ffmpeg executable. Default is `ffmpeg`, which means it
            should be in the system path.
        reportprogress : bool, default=False

        Yields
        ------
        Frames
            Iterator that generates numpy array frames (height x width x color
            channel).

        """
        for i,frame in enumerate(iterread_videofile(self.filepath,
                                                    startat=startat,
                                                    nframes=nframes,
                                                    color=color,
                                                    ffmpegpath=ffmpegpath)):
            if reportprogress:
                progress(i, self.nframes)
            yield frame

    def get_frame(self, framenumber, color=True, ffmpegpath='ffmpeg'):
        """Get frame specified by frame sequence number.

        Note that this can take a lot of processing because the video
        has to be decoded up to that number. Specifying by time is more
        efficient (see: `get_frameat` method.

        Parameters
        ----------
        framenumber : int
            Get the frame `framenumber` from the video stream.
        color : bool, default=True
            Read as a color frame (3 dimensional) or as a gray frame (2
            dimensional).
        ffmpegpath : str or pathlib.Path, optional
            Path to ffmpeg executable. Default is `ffmpeg`, which means it
            should be in the system path.

        Returns
        -------
        numpy ndarray
            The frame at the specified framenumber.

        Example
        -------
        >>> import birdwatcher as bw
        >>> vfs = bw.testvideosmall()
        >>> frame = vfs.get_frame(500)

        """
        return get_frame(self.filepath, framenumber=framenumber,
                         color=color, ffmpegpath=ffmpegpath)

    def get_frameat(self, time, color=True, ffmpegpath='ffmpeg'):
        """Get frame at specified time.

        Parameters
        ----------
        time : str
            Get frame at a time point in the video file. You
            can use two different time unit formats: sexagesimal
            (HOURS:MM:SS.MILLISECONDS, as in 01:23:45.678), or in seconds.
        color : bool, default=True
            Read as a color frame (2 dimensional) or as a gray frame (3
            dimensional).
        ffmpegpath : str or pathlib.Path, optional
            Path to ffmpeg executable. Default is `ffmpeg`, which means it
            should be in the system path.

        Returns
        -------
        numpy ndarray
            The frame at the specified time.

        Example
        -------
        >>> import birdwatcher as bw
        >>> vfs = bw.testvideosmall()
        >>> frame = vfs.get_frameat('5.05') # at 5 sec and 50 msec
        >>> frame = vfs.get_frameat('00:00:05.05') # same thing

        """
        return get_frameat(self.filepath, time=time, color=color,
                           ffmpegpath=ffmpegpath)

    def show(self, startat=None, nframes=None, framerate=None):
        """Shows frames in a video window.

        The frames of a VideoFileStream are displayed in a seperate window.
        Press 'q' to quit the video before the end.

        Parameters
        ----------
         startat : str, optional
            If specified, start at this time point in the video file. You
            can use two different time unit formats: sexagesimal
            (HOURS:MM:SS.MILLISECONDS, as in 01:23:45.678), or in seconds.
        nframes  : int, optional
            Read a specified number of frames.
        framerate : int, optional
            The default framerate is None, which means that the average frame
            rate of video stream, as reported in the metadata, is used.

        """
        if framerate is None:
            framerate = self.avgframerate
        f = self.iter_frames(startat=startat, nframes=nframes)
        return f.show(framerate=framerate)


def testvideosmall():
    """A 20-s video of a zebra finch for testing purposes.

    Returns
    -------
    VideoFileStream
        An instance of Birdwatcher's VideoFileSteam class.

    """
    file = 'zf20s_low.mp4'
    path = Path(__file__).parent / 'testvideos' / file
    return VideoFileStream(path)


def walk_videofiles(dirpath, extension='.avi'):
    """Walks recursively over contents of `dirpath` and yield pathlib Path
    objects of videofiles, as defined by their `extension`.

    Parameters
    ----------
    dirpath : str or Path
        The top-level directory to start at.
    extension : str, default='.avi'
        Filter on this extension.

    """

    dirpath = Path(dirpath)
    if extension.startswith('.'):
        extension = extension[1:]
    for file in dirpath.rglob(f'*.{extension}'):
        yield file


## FIXME collect much more info
def videofilesduration(dirpath, extension='avi'):
    files = sorted(
        [f for f in walk_videofiles(dirpath=dirpath, extension=extension)])
    s = 0
    for i, f in enumerate(files):
        s += VideoFileStream(f).duration


"""This module provides background subtractors. Background subtraction is a
major preprocessing step in many computer vision applications. It extracts the
moving foreground from static background.

All classes are based on OpenCV's background subtraction algorithms. They
can be used for example in movement detection. Note that OpenCV's API to
parameters of these algorithms is inconsistent. Sometimes parameters can be
provided at instantiation, sometimes you can change them with a method. In
Birdwatcher you can only provide parameters to background subtractor objects
at instantiation. The parameter names follow those of OpenCV to avoid
confusion.

"""
# TODO: incorporate the use of https://github.com/andrewssobral/bgslibrary

import cv2 as cv
import numpy as np


__all__ = ['BackgroundSubtractorMOG2', 'BackgroundSubtractorKNN',
           'BackgroundSubtractorLSBP']


class BaseBackgroundSubtractor:

    _initparams = {}                # to be implemented by subclass
    _setparams = {}                 # to be implemented by subclass
    _bgsubtractorcreatefunc = None  # to be implemented by subclass

    def __init__(self, **kwargs):
        # create background subtractor
        initkeys = set(kwargs.keys()) & set(self._initparams.keys())
        initparams = self._initparams.copy()
        for initkey in initkeys:
            initparams[initkey] = kwargs[initkey]
        self._bgs = self._bgsubtractorcreatefunc(**initparams)
        setkeys = set(kwargs.keys()) & set(self._setparams.keys())
        setparams = self._setparams.copy()
        for setkey in setkeys:
            setparams[setkey] = kwargs[setkey]
        self._set_params(**setparams)
        self._params = {**initparams, **setparams}

    def __str__(self):
        return self.__class__.__name__

    def __repr__(self):
        return f"<{self.__class__.__name__} {hex(id(self))}>"

    def _set_params(self, **kwargs):
        for key, val in kwargs.items():
            methodname = f'set{key}'
            self._bgs.__getattribute__(methodname)(val)

    def get_params(self):
        """Get parameters of the background subtraction algorithm.

        Returns
        -------
        dict
            Dictionary with algorithm parameters.

        """
        return self._params

    def apply(self, frame, fgmask=None, learningRate=-1.0):
        """Computes a foreground mask.

        Parameters
        ----------
        frame : numpy array image
            Next video frame.
        fgmask : numpy array image, optional
            The output foreground mask as an 8-bit binary image.
        learningRate : float, optional
            The value between 0 and 1 that indicates how fast the background
            model is learnt. The default negative parameter value (-1.0) makes
            the algorithm to use some automatically chosen learning rate. 0
            means that the background model is not updated at all, 1 means
            that the background model is completely reinitialized from the
            last frame.

        Returns
        -------
        image frame
            The output foreground mask as an 8-bit image.

        """

        return self._bgs.apply(image=frame, fgmask=fgmask,
                               learningRate=learningRate)


class BackgroundSubtractorKNN(BaseBackgroundSubtractor):
    """Wraps OpenCV's `BackgroundSubtractorKNN` class. Parameter names follow
    those in OpenCV.

    Parameters
    ----------
    History : int, default=5
        Length of the history.
    kNNSamples : int, default=10
        The number of neighbours, the k in the kNN. K is the number of
        samples that need to be within dist2Threshold in order to decide
        that that pixel is matching the kNN background model.
    NSamples : int, default=6
        The number of data samples in the background model.
    Dist2Threshold : float, default=500
        Threshold on the squared distance between the pixel and the sample
        to decide whether a pixel is close to that sample. This parameter
        does not affect the background update.
    DetectShadows : bool, default=False
        If true, the algorithm detects shadows and marks them.
    ShadowThreshold : float, default=0.5
        A shadow is detected if pixel is a darker version of the background.
        The shadow threshold is a threshold defining how much darker the
        shadow can be. 0.5 means that if a pixel is more than twice darker
        then it is not shadow.
    ShadowValue : int, default=127
        Shadow value is the value used to mark shadows in the foreground mask.
        Value 0 in the mask always means background, 255 means foreground.

    """
    
    _setparams = {'History': 5,
                  'kNNSamples': 10,
                  'NSamples': 6,
                  'Dist2Threshold': 500,
                  'DetectShadows': False,
                  'ShadowThreshold': 0.5,
                  'ShadowValue': 127}

    _bgsubtractorcreatefunc = cv.createBackgroundSubtractorKNN


class BackgroundSubtractorMOG2(BaseBackgroundSubtractor):
    """Wraps OpenCV's `BackgroundSubtractorMOG2` class. Parameter names follow
    those in OpenCV.

    Parameters
    ----------
    History : int, default=3
        Length of the history.
    ComplexityReductionThreshold : float, default=0.5
        This parameter defines the number of samples needed to accept to prove
        the component exists. CT=0.05 is a default value for all the samples.
        By setting CT=0 you get an algorithm very similar to the standard
        Stauffer&Grimson algorithm.
    BackgroundRatio : float, default=0.1
        If a foreground pixel keeps semi-constant value for about
        backgroundRatio*history frames, it's considered background and added
        to the model as a center of a new component. It corresponds to TB
        parameter in the paper.
    NMixtures : int, default=7
        The number of gaussian components in the background model.
    VarInit : int, default=15
        The initial variance of each gaussian component.
    VarMin : int, default=10
        The minimum variance of each gaussian component.
    VarMax : int, default=75
        The maximum variance of each gaussian component.
    VarThreshold : int, default=70
        The variance threshold for the pixel-model match. The main threshold
        on the squared Mahalanobis distance to decide if the sample is well
        described by the background model or not. Related to Cthr from the
        paper.
    VarThresholdGen : int, default=9
        The variance threshold for the pixel-model match used for new mixture
        component generation. Threshold for the squared Mahalanobis distance
        that helps decide when a sample is close to the existing components
        (corresponds to Tg in the paper). If a pixel is not close to any
        component, it is considered foreground or added as a new component.
        3 sigma => Tg=3*3=9 is default. A smaller Tg value generates more
        components. A higher Tg value may result in a small number of
        components but they can grow too large.
    DetectShadows : bool, default=False
        If true, the algorithm detects shadows and marks them.
    ShadowThreshold : float, default=0.5
        A shadow is detected if pixel is a darker version of the background.
        The shadow threshold is a threshold defining how much darker the
        shadow can be. 0.5 means that if a pixel is more than twice darker
        then it is not shadow.
    ShadowValue : int, default=0
        Shadow value is the value used to mark shadows in the foreground mask.
        Value 0 in the mask always means background, 255 means foreground.

    """
    
    _setparams = {'History': 3,
                  'ComplexityReductionThreshold': 0.05,
                  'BackgroundRatio': 0.1,
                  'NMixtures': 7,
                  'VarInit': 15,
                  'VarMin': 10,
                  'VarMax': 75,
                  'VarThreshold': 70,
                  'VarThresholdGen': 9,
                  'DetectShadows': False,
                  'ShadowThreshold': 0.5,
                  'ShadowValue': 0}

    _bgsubtractorcreatefunc = cv.createBackgroundSubtractorMOG2

#TODO this bgs does nor work with roi or nroi
class BackgroundSubtractorLSBP(BaseBackgroundSubtractor):
    """Wraps OpenCV's `BackgroundSubtractorLSBP` class. Parameter names follow
    those in OpenCV.

    Parameters
    ----------
    mc : int, default=0
        Whether to use camera motion compensation.
    nSamples : int, default=20
        Number of samples to maintain at each point of the frame.
    LSBPRadius : int, default=16
        LSBP descriptor radius.
    Tlower : float, default=2.0
        Lower bound for T-values. See [103] for details.
    Tupper : float, default=32.0
        Upper bound for T-values. See [103] for details.
    Tinc : float, default=1.0
        Increase step for T-values.
    Tdec : float, default=0.05
        Decrease step for T-values.
    Rscale : float, default=10.0
        Scale coefficient for threshold values.
    Rincdec : float, default=0.005
        Increase/Decrease step for threshold values.
    noiseRemovalThresholdFacBG : float, default=0.0004
        Strength of the noise removal for background points.
    noiseRemovalThresholdFacFG : float, default=0.0008
        Strength of the noise removal for foreground points.
    LSBPthreshold : int, default=8
        Threshold for LSBP binary string.
    minCount : int, default=2
        Minimal number of matches for sample to be considered as foreground.

    """

    _initparams = {'mc': 0,
                   'nSamples': 20,
                   'LSBPRadius': 16,
                   'Tlower': 2.0,
                   'Tupper': 32.0,
                   'Tinc': 1.0,
                   'Tdec': 0.05,
                   'Rscale': 10.0,
                   'Rincdec': 0.005,
                   'noiseRemovalThresholdFacBG': 0.0004,
                   'noiseRemovalThresholdFacFG': 0.0008,
                   'LSBPthreshold': 8,
                   'minCount': 2}

    _bgsubtractorcreatefunc = cv.bgsegm.createBackgroundSubtractorLSBP


import unittest
import tempfile
import shutil

import numpy as np
from numpy.testing import assert_array_equal

import birdwatcher as bw
from birdwatcher.frames import Frames, FramesColor, framecolor, framegray


colorlist = [framecolor(height=480, width=640, color=(0,0,0)),
             framecolor(height=480, width=640, color=(1,1,1))]

graylist = [framegray(height=480, width=640, value=0),
            framegray(height=480, width=640, value=1)]


class TestFrames(unittest.TestCase):

    def test_size(self):
        frames = Frames(colorlist)
        self.assertEqual(frames.frameheight, 480)
        self.assertEqual(frames.framewidth, 640)

    def test_color(self):
        for frame in FramesColor(5, height=480, width=640, color= (0,0,0)):
            self.assertEqual(frame.sum(), 0)
    
    def test_calcmeanframe(self):
        frames = Frames([framecolor(height=480, width=640, color=(10,10,10)),
                         framecolor(height=480, width=640, color=(30,30,30))])
        meanframe = frames.calc_meanframe()
        self.assertTupleEqual(meanframe.shape, (480, 640, 3))
        self.assertEqual(meanframe[0,0].sum(), 60)


class TestPeekFrame(unittest.TestCase):

    def test_peekframe(self):
        frames = Frames(colorlist)
        frame0 = frames.peek_frame()
        self.assertEqual(frame0.sum(), colorlist[0].sum())
        outputframes = [frame for frame in frames]
        self.assertEqual(len(outputframes), 2) # is the first frame still available?
        self.assertEqual(outputframes[1].sum(), colorlist[1].sum())


class TestFrameIterator(unittest.TestCase):

    def test_drawframenumber(self):
        frames = FramesColor(5, height=480, width=640, color=(0,0,0))
        for frame in frames.draw_framenumbers():
            self.assertGreater(frame.sum(), 0)

    def test_edgedetection(self):
        frames = Frames(graylist).edge_detection()
        self.assertIsInstance(frames, Frames)
        outputframes = [frame for frame in frames]
        self.assertEqual(len(outputframes), 2)
        self.assertTupleEqual(graylist[0].shape, outputframes[0].shape)


class TestFindNonZero(unittest.TestCase):
    
    def setUp(self):
        self.fgray = framegray(width=640, height=480)
        self.fgray[1,1] = 1
        self.fgray[1,2] = 2
        self.fcolor = framecolor(width=640, height=480, color=(0, 0, 0))
        self.fcolor[1,1] = (1, 1, 1)
        self.fcolor[1,2] = (1, 0, 0)
        self.fcolor[1,3] = (254, 1, 1) # sums to 0 if uint8 sum

    def test_grayframe(self):
        idx = Frames([self.fgray]).find_nonzero()
        assert_array_equal(next(idx), np.array([[1,1], [2,1]] ,dtype='int32'))

    def test_colorframe(self):
        idx = Frames([self.fcolor]).find_nonzero()
        assert_array_equal(next(idx), np.array([[1,1], [2,1], [3,1]], 
                                               dtype='int32'))
    def test_saveascoords(self):
        frames = Frames([self.fcolor, self.fcolor])
        tempdirname = tempfile.mkdtemp()        
        coordsarray = frames.save_nonzero(filepath=tempdirname, 
                                          metadata={'avgframerate': 5})
        self.assertIsInstance(coordsarray, bw.CoordinateArrays)
        shutil.rmtree(tempdirname)


import unittest
import tempfile
import shutil
from pathlib import Path

import pandas as pd
import matplotlib.pyplot as plt
import seaborn as sns

import birdwatcher as bw
import birdwatcher.movementdetection as md


settings = {'bgs_params':  {'History': [4, 12],
                            'ComplexityReductionThreshold': [0.05],
                            'BackgroundRatio': [0.1],
                            'NMixtures': [7],
                            'VarInit': [15],
                            'VarMin': [4],
                            'VarMax': [75],
                            'VarThreshold': [10],
                            'VarThresholdGen': [9],
                            'DetectShadows': [False],
                            'ShadowThreshold': [0.5],
                            'ShadowValue': [0]},
            'processing':  {'color': [True],
                            'resizebyfactor': [1],
                            'blur': [0,],
                            'morphologyex': [True]}}


class TestParameterSelection(unittest.TestCase):

    def setUp(self):
        self.tempdirname1 = Path(tempfile.mkdtemp())
        try:
            self.params = md.apply_all_parameters(bw.testvideosmall(), 
                                                  settings, nframes=200)
        except:
            self.tearDown()
            raise

    def tearDown(self):
        shutil.rmtree(self.tempdirname1)

    def test_attributes(self):
        self.assertIsInstance(self.params, md.ParameterSelection)
        self.assertIsInstance(self.params.df, pd.DataFrame)
        self.assertIsInstance(self.params.vfs, bw.VideoFileStream)
        self.assertEqual(self.params.bgs_type, 
                         str(bw.BackgroundSubtractorMOG2))
        self.assertEqual(self.params.startat, None)
        self.assertEqual(self.params.nframes, 200)
        self.assertEqual(self.params.path, None)

    def test_getvideofragment(self):
        frames = self.params.get_videofragment()
        self.assertIsInstance(frames, bw.Frames)

    def test_getparams(self):
        all = self.params.get_parameters('all')
        settings_flat = {**settings['bgs_params'], **settings['processing']}
        self.assertEqual(all, settings_flat)
        multi_only = self.params.get_parameters('multi_only')
        self.assertEqual(multi_only, {'History': [4, 12]})

    def test_plotparams(self):
        g = self.params.plot_parameters(rows='History', cols=None, 
                                        default_values={'History': 4})
        self.assertIsInstance(g, sns.axisgrid.FacetGrid)
        self.assertEqual(g.axes.shape, (2,1))
        plt.close(g.figure)

    def test_drawcircles(self):
        frames, colorspecs = self.params.draw_multiple_circles({'History':
                                                                [4,12]})
        self.assertIsInstance(frames, bw.Frames)
        self.assertEqual(colorspecs.shape, (1,2))

    def test_saveparams(self):
        self.assertIsNone(self.params.path)
        self.params.save_parameters(self.tempdirname1)
        self.assertIsNotNone(self.params.path)

    def test_saveanotherparams(self):
        self.params.save_parameters(self.tempdirname1)
        self.params.save_parameters(self.tempdirname1)
        self.assertEqual(self.params.path[-1], '2')


class TestParamSelectionCount(unittest.TestCase):

    def setUp(self):
        self.tempdirname1 = Path(tempfile.mkdtemp())
        try:
            self.params = md.apply_all_parameters(bw.testvideosmall(), 
                                                  settings, nframes=200,
                                                  use_stats='count')
        except:
            self.tearDown()
            raise

    def tearDown(self):
        shutil.rmtree(self.tempdirname1)

    def test_count(self):
        coordslabels = self.params.df.coords.unique()
        self.assertEqual(coordslabels.shape, (1,))
        self.assertEqual(coordslabels[0], 'count')


import unittest
import tempfile
import shutil
from pathlib import Path

import birdwatcher as bw
import birdwatcher.movementdetection as md


settings = {'bgs_params':  {'History': 12,
                            'ComplexityReductionThreshold': 0.05,
                            'BackgroundRatio': 0.1,
                            'NMixtures': 7,
                            'VarInit': 15,
                            'VarMin': 10,
                            'VarMax': 75,
                            'VarThreshold': 70,
                            'VarThresholdGen': 9,
                            'DetectShadows': False,
                            'ShadowThreshold': 0.5,
                            'ShadowValue': 0},
            'processing':  {'color': False,
                            'resizebyfactor': 1,
                            'blur': 10,
                            'morphologyex': True}}


class TestApplySettingst(unittest.TestCase):

    def test_applysettings(self):
        vfs = bw.testvideosmall()
        settings_flat = {**settings['bgs_params'], **settings['processing']}
        frames = md.apply_settings(vfs, settings_flat)
        self.assertIsInstance(frames, bw.Frames)


class TestDetectMovement(unittest.TestCase):

    def setUp(self):
        self.tempdirname1 = Path(tempfile.mkdtemp())
        self.vfs = bw.testvideosmall()

    def tearDown(self):
        shutil.rmtree(self.tempdirname1)

    def test_detectmovement(self):
        cd, _, _ = md.detect_movement(self.vfs, nframes=200,
                                      analysispath=self.tempdirname1,
                                      overwrite=True)
        self.assertIsInstance(cd, bw.CoordinateArrays)

    def test_movementsettings(self):
        cd, _, _ = md.detect_movement(self.vfs, settings=settings, 
                                      nframes=200, 
                                      analysispath=self.tempdirname1, 
                                      overwrite=True)
        self.assertEqual(cd.metadata['settings']['History'], 12)
        self.assertEqual(cd.metadata['settings']['blur'], 10)
    
    def test_exception(self):
        with self.assertRaises(TypeError):
            md.detect_movement('not_a_videofilestream_object')


class TestBatchDetectMovement(unittest.TestCase):

    def setUp(self):
        self.tempdirname1 = Path(tempfile.mkdtemp())
        self.vfs1 = (bw.testvideosmall()
                    .iter_frames(nframes=200)
                    .tovideo(self.tempdirname1 / 'video1.mp4', framerate=25))
        self.vfs2 = (bw.testvideosmall()
                    .iter_frames(nframes=200)
                    .tovideo(self.tempdirname1 / 'video2.mp4', framerate=25))

    def tearDown(self):
         shutil.rmtree(self.tempdirname1)

    def test_batchdetection(self):
        md.batch_detect_movement([self.vfs1,self.vfs2], nframes=200,
                                 analysispath=self.tempdirname1, 
                                 overwrite=True, archived=False)
        self.assertTrue(Path.exists(self.tempdirname1/'movement_video1/coords.darr'))
        self.assertTrue(Path.exists(self.tempdirname1/'movement_video2/coords.darr'))

    def test_batcharchive(self):
        md.batch_detect_movement([self.vfs1,self.vfs2], nframes=200,
                                 analysispath=self.tempdirname1,
                                 overwrite=True, archived=True)
        self.assertTrue(Path.exists(self.tempdirname1/'movement_video1/coords.darr.tar.xz'))
        self.assertTrue(Path.exists(self.tempdirname1/'movement_video2/coords.darr.tar.xz'))


import sys
import setuptools

# we follow https://scientific-python.org/specs/spec-0000/
if sys.version_info < (3,10):
    print("Birdwatcher requires Python 3.11 or higher please upgrade")
    sys.exit(1)


setuptools.setup(
    packages=['birdwatcher', 'birdwatcher.movementdetection', 'birdwatcher.tests', 'birdwatcher.testvideos'],
    package_data={'birdwatcher.testvideos': ['*.mp4']},
    include_package_data=True,
)



"""This module provides an interface for video file input and output using
FFmpeg. Most users will not use functions and classes from this module
directly, but rather through methods and functions in other modules. We use
FFmpeg instead of video IO in OpenCV because we want to be more flexible so
that the user can decide for a (potentially self-compiled) version of FFmpeg
that supports functionality not present in OpenCV. E.g., specific codecs,
CUDA support. It is easier to compile or find specific versions of FFmpeg
than of OpenCV. Keeping all video IO in this module, allows for the easy
addition of other ways of video IO in Birdwatcher in the future.

"""

import json
import subprocess
from pathlib import Path

import numpy as np

from .utils import peek_iterable


__all__ = ['arraytovideo']


class FFmpegError(Exception):
    def __init__(self, cmd, stdout, stderr):
        super().__init__(f'{cmd} error (see stderr output for '
                                    f'detail)')
        self.stdout = stdout
        self.stderr = stderr


def arraytovideo(frames, filepath, framerate, scale=None, crf=17,
                 format='mp4', codec='libx264', pixfmt='yuv420p',
                 ffmpegpath='ffmpeg', loglevel='quiet'):
    """Writes an iterable of numpy frames as video file using ffmpeg.

    Parameters
    ----------
    frames : iterable
        Iterable should produce numpy height x width x channel arrays with
        values ranging from 0 to 255. Frames can be color (3-dim) or gray
        (2-dim).
    filepath : str
        Name of the videofilepath that should be written to.
    framerate : int
        Framerate of video in frames per second.
    scale : tuple, optional
        (width, height). The default (None) does not change width and height.
    crf : int, default=17
        Value determines quality of video. The default 17 is high quality.
        Use 23 for good quality.
    format : str, default='mp4'
        ffmpeg video format.
    codec : str, default='libx264'
        ffmpeg video codec.
    pixfmt : str, default='yuv420p'
        ffmpeg pixel format.
    ffmpegpath : str or pathlib.Path, optional
        Path to ffmpeg executable. Default is `ffmpeg`, which means it
        should be in the system path.
    loglevel : {'quiet', 'panic', 'fatal', 'error', 'warning', 'info',
               'verbose', 'debug' ,'trace'}, optional
        Level of info that ffmpeg should print to terminal. Default is
        'quiet'.

    Notes
    -----
    See ffmpeg documentation for more information.

    """
    _check_loglevelarg(loglevel)
    frame, framegen = peek_iterable(frames)
    height, width, *_ = frame.shape
    Path(filepath).parent.mkdir(parents=True, exist_ok=True)   
    if frame.ndim == 2:
        ipixfmt = 'gray'
    else: # frame.ndim == 3:
        ipixfmt = 'bgr24'

    args = [str(ffmpegpath),
            #'-hwaccel',
            '-loglevel' , loglevel,
            '-f', 'rawvideo',
            '-vcodec','rawvideo',
            '-pix_fmt', ipixfmt,
            '-r', f'{framerate}',
            '-s', f'{width}x{height}', '-i', 'pipe:']
    if codec is not None:
        args += ['-vcodec', f'{codec}']
    if format is not None:
        args += ['-f', f'{format}']
    if crf is not None:
        args += ['-crf', f'{crf}']
    if pixfmt is not None:
        args +=['-pix_fmt', f'{pixfmt}']
    if scale is not None:
        width, height = scale
        args.extend(['-vf', f'scale={width}:{height}'])
    args.extend([filepath, '-y'])
    p = subprocess.Popen(args, stdin=subprocess.PIPE, stderr=subprocess.PIPE,
                         stdout=subprocess.PIPE)
    for frame in framegen:
        # if frame.ndim == 2:
        #     frame = cv.cvtColor(frame, cv.COLOR_GRAY2BGR)
        p.stdin.write(frame.astype(np.uint8).tobytes())
    out, err = p.communicate()
    p.stdin.close()
    
    return Path(filepath)


def videofileinfo(filepath, ffprobepath='ffprobe'):
    args = [str(ffprobepath), '-print_format', 'json', '-show_format',
            '-show_streams', str(filepath)]
    p = subprocess.Popen(args, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    return json.loads(p.stdout.read().decode('utf-8'))

## FIXME inform before raising StopIteration that file has no frames
def iterread_videofile(filepath, startat=None, nframes=None, color=True,
                       ffmpegpath='ffmpeg', loglevel='quiet'):
    """
    Parameters
    ----------
    filepath : str
        Name of the videofilepath.
    startat : str
      There are two accepted syntaxes for expressing time duration.
      [-][HH:]MM:SS[.m...], where HH expresses the number of hours,
      MM the number of minutes for a maximum of 2 digits, and SS
      the number of seconds for a maximum of 2 digits. The m at the
      end expresses decimal value for SS.
      [-]S+[.m...][s|ms|us], where S expresses the number of seconds,
      with the optional decimal part m. The optional literal suffixes
      ‘s’, ‘ms’ or ‘us’ indicate to interpret the value as seconds,
      milliseconds or microseconds, respectively. The following
      examples are all valid time duration: ‘55’ means 55 seconds,
      ‘0.2’ means 0.2 seconds, ‘200ms’ means 200 milliseconds,
      ‘200000us’ means 200000 microseconds, ‘12:03:45’ means 12 hours,
      03 minutes and 45 seconds, ‘23.189’ means 23.189 seconds.
    nframes : int
    color : bool, default=True
    ffmpegpath : str or pathlib.Path, optional
        Path to ffmpeg executable. Default is `ffmpeg`, which means it
        should be in the system path.
    loglevel : {'quiet', 'panic', 'fatal', 'error', 'warning', 'info',
               'verbose', 'debug' , 'trace'}, optional
        Level of info that ffmpeg should print to terminal. Default is
        'quiet'.

    Yields
    -------
    numpy ndarray
        Generates numpy arrays of video frames.

    """
    _check_loglevelarg(loglevel)
    frameshape, framesize, frameheight, framewidth, pix_fmt = \
        _get_frameproperties(filepath=filepath, color=color)
    args = [str(ffmpegpath), '-loglevel' , loglevel]
    if startat is not None:
        args.extend(['-ss', startat, '-i', str(filepath)])
    else:
        args.extend(['-i', str(filepath)])
    if nframes is not None:
        args += ['-vframes', str(nframes)]
    args +=['-vcodec', 'rawvideo', '-pix_fmt', pix_fmt,
            '-f', 'rawvideo', 'pipe:1']
    with subprocess.Popen(args, stdout=subprocess.PIPE,
                          stderr=None) as p:
        frameno = 0
        while True:
            data = p.stdout.read(framesize)
            ar = np.frombuffer(data, dtype=np.uint8)
            if (ar.size == framesize) and ((nframes is None) or (frameno <
                                                                 nframes)):
                yield ar.reshape(frameshape)
                frameno += 1
            else:
                break

#TODO check threads code
def count_frames(filepath, threads=8, ffprobepath='ffprobe'):
    args = [str(ffprobepath), '-threads:0', str(threads),
            '-count_frames', '-select_streams', 'v:0', '-show_entries',
            'stream=nb_read_frames', '-print_format', 'json', str(filepath)]
    p = subprocess.Popen(args, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    out = json.loads(p.stdout.read().decode('utf-8'))
    err = p.stderr.read().decode('utf-8')
    if not out:
        raise FFmpegError(ffprobepath, out, err)
    return int(out['streams'][0]['nb_read_frames'])

# def get_frame_old(filepath, framenumber, color=True, ffmpegpath='ffmpeg'):
#     for frame in iterread_videofile(filepath, startat=None, nframes=framenumber+1,
#                                     color=color, ffmpegpath=ffmpegpath):
#         pass
#     return frame


def get_frame(filepath, framenumber, color=True, ffmpegpath='ffmpeg',
              loglevel= 'quiet'):
    _check_loglevelarg(loglevel)
    frameshape, framesize, frameheight, framewidth, pix_fmt = \
        _get_frameproperties(filepath=filepath, color=color)
    args = [str(ffmpegpath), '-loglevel' , loglevel, '-i', str(filepath)]
    args +=['-vcodec', 'rawvideo',  '-vf', f"select='eq(n\\,{framenumber})'",
            '-vframes', '1', '-pix_fmt', pix_fmt,
            '-f', 'rawvideo', 'pipe:1']
    with subprocess.Popen(args, stdout=subprocess.PIPE,
                          stderr=subprocess.PIPE) as p:
        return np.frombuffer(p.stdout.read(framesize),
                             dtype=np.uint8).reshape(frameshape)


def get_frameat(filepath, time, color=True, ffmpegpath='ffmpeg', 
                loglevel='quiet'):
    return next(iterread_videofile(filepath, startat=time, nframes=1, 
                                   color=color, ffmpegpath=ffmpegpath, 
                                   loglevel=loglevel))


def extract_audio(filepath, outputpath=None, overwrite=False, 
                  codec='pcm_s24le', channel=None, ffmpegpath='ffmpeg', 
                  loglevel='quiet'):
    """Extract audio as wav file.

    Parameters
    ----------
    outputpath : str or pathlib.Path, optional
        Filename and path to write audio to. The default is None, which means 
        the same directoy and name as the video file is used, but then with 
        '.wav' extension.
    overwrite : bool, default=False
        Overwrite if audio file exists or not.
    codec : str, default='pcm_s24le'
        ffmpeg audio codec, with 24-bit pcm as default output.
    channel : int, default=None
        Channel number to extract. The default None will extract all channels.
    ffmpegpath : str or pathlib.Path, optional
        Path to ffmpeg executable. Default is `ffmpeg`, which means it should 
        be in the system path.
    loglevel : {'quiet', 'panic', 'fatal', 'error', 'warning', 'info', 
                'verbose', 'debug' ,'trace'}, optional
        Level of info that ffmpeg should print to terminal. Default is 
        'quiet'.

    """
    filepath = Path(filepath)
    if outputpath is None:
        outputpath = filepath.with_suffix('.wav')
    else:
        outputpath = Path(outputpath)
    if outputpath.exists() and not overwrite:
        raise IOError(f'"{outputpath}" already exists, use overwrite parameter')
    _check_loglevelarg(loglevel)
    args = [str(ffmpegpath), '-loglevel' , loglevel, '-y',
            '-i', str(filepath),
            '-vn',
            '-codec:a', codec]
    if channel is not None:
        args += ['-af', f'pan=mono|c0=c{channel-1}']
    args += [str(outputpath)]
    with subprocess.Popen(args, stdout=subprocess.PIPE,
                          stderr=subprocess.PIPE) as p:
        out, err = p.communicate()
    if err:
        return err.decode('utf-8')


def _get_frameproperties(filepath, color):
    """Convenience function that produces frame characteristics for a given
    video. Handy if you want to know the format of a frame that is returned
    by ffmpeg from a pipe."""
    vfi = videofileinfo(filepath)
    frameheight = vfi['streams'][0]['height']
    framewidth = vfi['streams'][0]['width']
    if color:
        frameshape = (frameheight, framewidth, 3)
        framesize = frameheight * framewidth * 3
        pix_fmt = 'bgr24'
    else:
        frameshape = (frameheight, framewidth)
        framesize = frameheight * framewidth
        pix_fmt = 'gray'
    return frameshape, framesize, frameheight, framewidth, pix_fmt


def _check_loglevelarg(loglevelarg):
    levels = ('quiet', 'panic', 'fatal', 'error', 'warning', 'info',
              'verbose', 'debug', 'trace')
    if loglevelarg not in levels:
        raise ValueError(f"`loglevel` argument ('f{loglevelarg}') "
                         f"should be one of: {levels}")


from pathlib import Path

import numpy as np
import darr

import birdwatcher as bw
from birdwatcher.utils import derive_filepath
from birdwatcher.coordinatearrays import _archive


__all__ = ['batch_detect_movement', 'detect_movement', 'apply_settings', 
           'create_movementvideo']


default_settings = {'processing':{'color': False, # booleans only
                                  'blur': 0, # use '0' for no blur
                                  'morphologyex': True, # booleans only
                                  'resizebyfactor': 1}} # use '1' for no change in size


def batch_detect_movement(vfs_list, settings=None, startat=None, 
                          nframes=None, roi=None, nroi=None, 
                          bgs_type=bw.BackgroundSubtractorMOG2, 
                          analysispath='.', ignore_firstnframes=10, 
                          overwrite=False, resultvideo=False, 
                          archived=True, nprocesses=6):
    """Batch function to apply movement detection to a list of videofiles.

    As default, the coordinate arrays are saved as archived data folders to 
    reduce memory use. Compression of coordinate results takes a long time and 
    is single-threaded. We therefore do this in parallel. Use the `nprocesses` 
    parameter to specify the number of cores devoted to this.

    Parameters
    ----------
    archived : bool, default=True
        As default, the coordinate data are saved as archived data folders. 
        Use False, to save the coordinate data without compression.
    nprocesses : int, default=6
        The number of cores devoted to archiving the coordinate data, if 
        `archived` is set `True`.

    See the docstrings of `dectect_movement` for an explanation of the other 
    parameters.

    """
    from multiprocessing.pool import ThreadPool

    tobearchived = []
    for i, vfs in enumerate(vfs_list):
        cd, cc, cm = detect_movement(vfs, settings=settings, startat=startat, 
                                     nframes=nframes, roi=roi, nroi=nroi, 
                                     bgs_type=bgs_type, 
                                     analysispath=analysispath, 
                                     ignore_firstnframes=ignore_firstnframes, 
                                     overwrite=overwrite, 
                                     resultvideo=resultvideo)
        if archived:
            tobearchived.append(cd)
            if (len(tobearchived) == nprocesses) or (i == (len(vfs_list)-1)):
                with ThreadPool(processes=nprocesses) as pool:
                    list([i for i in pool.imap_unordered(_archive, tobearchived)])
                tobearchived = []


def detect_movement(vfs, settings=None, startat=None, nframes=None, roi=None, 
                    nroi=None, bgs_type=bw.BackgroundSubtractorMOG2, 
                    analysispath='.', ignore_firstnframes=10, 
                    overwrite=False, resultvideo=False):
    """Detects movement based on a background subtraction algorithm.

    High-level function to perform movement detection with default parameters, 
    but also the option to modify many parameter settings.

    Parameters
    ----------
    vfs : VideoFileStream
        A Birdwatcher VideoFileStream object.
    settings : {dict, dict}, optional
        Dictionary with two dictionaries. One 'bgs_params' with the parameter 
        settings from the BackgroundSubtractor and another 'processing' 
        dictionary with settings for applying color, resizebyfactor, blur and 
        morphologyex manipulations. If None, the default settings of the 
        BackgroundSubtractor are used on grey color frames, including 
        morphological transformation to reduce noise.
    startat : str, optional
        If specified, start at this time point in the video file. You can use 
        two different time unit formats: sexagesimal 
        (HOURS:MM:SS.MILLISECONDS, as in 01:23:45.678), or in seconds.
    nframes  : int, optional
            Read a specified number of frames.
    roi : (int, int, int, int), optional
        Region of interest. Only look at this rectangular region. h1, h2, w1, 
        w2.
    nroi : (int, int, int, int), optional
        Not region of interest. Exclude this rectangular region. h1, h2, w1, 
        w2.
    bgs_type: BackgroundSubtractor, default=bw.BackgroundSubtractorMOG2
        This can be any of the BackgroundSubtractors in Birdwatcher, e.g. 
        BackgroundSubtractorMOG2, BackgroundSubtractorKNN, 
        BackgroundSubtractorLSBP.
    analysispath : Path or str, optional
        Where to write results to. The default writes to the current working 
        directory.
    ignore_firstnframes : int, default=10
        Do not provide coordinates for the first n frames. These often have a 
        lot of false positives.
    overwrite : bool, default=False
        Overwrite results or not.
    resultvideo : bool, default=False
        Automatically generate a video with results, yes or no.

    Returns
    -------
    tuple of arrays (coordinates, coordinate count, coordinate mean)
        These are Darr arrays that are disk-based.

    """
    if not isinstance(vfs, bw.VideoFileStream):
        raise TypeError(f"`vfs` parameter not a VideoFileStream object.")

    output_settings = {**bgs_type().get_params(), 
                       **default_settings['processing']} # get flat dict
    if settings is not None:
        settings = {**settings['bgs_params'], **settings['processing']}
        output_settings.update(settings)

    movementpath = Path(analysispath) / f'movement_{vfs.filepath.stem}'
    Path(movementpath).mkdir(parents=True, exist_ok=True)

    metadata = {}
    metadata['backgroundsegmentclass'] = str(bgs_type)
    metadata['settings'] = output_settings
    metadata['startat'] = startat
    metadata['nframes'] = nframes
    metadata['roi'] = roi
    metadata['nroi'] = nroi
    metadata['birdwatcherversion'] = bw.__version__

    frames = apply_settings(vfs, output_settings, startat, nframes, roi, nroi, 
                            bgs_type)

    cd = frames.save_nonzero(Path(movementpath) / 'coords.darr',
                             metadata = metadata,
                             ignore_firstnframes = ignore_firstnframes,
                             overwrite = overwrite)
    cc = darr.asarray(Path(movementpath) / 'coordscount.darr', 
                      cd.get_coordcount(),
                      metadata=metadata, overwrite=True)
    cm = darr.asarray(Path(movementpath) / 'coordsmean.darr', 
                      cd.get_coordmean(),
                      metadata=metadata, overwrite=True)

    if resultvideo:
        create_movementvideo(vfs, cd, startat=startat, nframes=nframes,
                             videofilepath=Path(movementpath) / 
                             'movementvideo.mp4')
    return cd, cc, cm


def apply_settings(vfs, settings, startat=None, nframes=None, roi=None, 
                   nroi=None, bgs_type=bw.BackgroundSubtractorMOG2):
    """Applies movement detection based on various parameter settings.

    The background subtractor should be provided as a parameter.

    Parameters
    ----------
    vfs : VideoFileStream
        A Birdwatcher VideoFileStream object.
    settings : dict
        Dictionary with parameter settings from the BackgroundSubtractor and 
        settings for applying color, resizebyfactor, blur and morphologyex 
        manipulations.
    startat : str, optional
        If specified, start at this time point in the video file. You can use 
        two different time unit formats: sexagesimal 
        (HOURS:MM:SS.MILLISECONDS, as in 01:23:45.678), or in seconds.
    nframes  : int, optional
            Read a specified number of frames.
    roi : (int, int, int, int), optional
        Region of interest. Only look at this rectangular region. h1, h2, w1, 
        w2.
    nroi : (int, int, int, int), optional
        Not region of interest. Exclude this rectangular region. h1, h2, w1, 
        w2.
    bgs_type: BackgroundSubtractor, default=bw.BackgroundSubtractorMOG2
        This can be any of the BackgroundSubtractors in Birdwatcher, e.g. 
        BackgroundSubtractorMOG2, BackgroundSubtractorKNN, 
        BackgroundSubtractorLSBP.

    Yields
    ------
    Frames
        Iterator that generates numpy array frames (height x width x color 
        channel).

    """
    frames = vfs.iter_frames(startat=startat, nframes=nframes, 
                             color=settings['color'])

    if settings['resizebyfactor'] != 1:
        val = settings['resizebyfactor']
        frames = frames.resizebyfactor(val,val)
        
        # resizebyfactor roi and nroi
        if roi is not None:
            roi = tuple([int(r*val) for r in roi])
        if nroi is not None:
            nroi = tuple([int(r*val) for r in nroi])

    if settings['blur']:
        val = settings['blur']
        frames = frames.blur((val,val))

    bgs_params = bgs_type().get_params()
    bgs_params.update((k, v) for k, v in settings.items() if k in bgs_params)
    bgs = bgs_type(**bgs_params)

    frames = frames.apply_backgroundsegmenter(bgs, learningRate=-1, 
                                              roi=roi, nroi=nroi)
    if settings['morphologyex']:
        frames = frames.morphologyex(morphtype='open', kernelsize=2)

    return frames


def create_movementvideo(vfs, coords, startat=None, nframes=None, 
                         videofilepath=None, draw_mean=True, 
                         draw_framenumbers=(2, 25), crf=17, scale=None):
    """Create a nice video from the original video with movement detection
    results superimposed.
    
    Note that the parameters 'startat' and 'nframes' can be used to select a 
    fragment of the video, but these parameters are not used to get the same 
    selection of the movement detection results('coords'). Thus, always all 
    frames from 'coords' are used to plot on the video(fragment).

    Parameters
    ----------
    vfs : VideoFileStream
        A Birdwatcher VideoFileStream object
    coords : CoordinateArrays
        CoordinateArrays object with movement results.
    startat : str, optional
        If specified, start at this time point in the video file. You can use 
        two different time unit formats: sexagesimal 
        (HOURS:MM:SS.MILLISECONDS, as in 01:23:45.678), or in seconds.
    nframes  : int, optional
            Read a specified number of frames.
    videofilepath : Path or str, optional
        Output path. If None, writes to filepath of videofilestream.
    draw_mean : bool, default=True
        Draw the mean of the coordinates per frame, or not.
    draw_framenumbers : tuple, optional
        Draw frame numbers. A tuple of ints indicates where to draw
        them. The default (2, 25) draws numbers in the top left corner.
        To remove the framenumbers use None.
    crf : int, default=17
        Quality factor output video for ffmpeg. The default 17 is high
        quality. Use 23 for good quality.
    scale : tuple, optional
        (width, height). The default (None) does not change width and height.

    Returns
    -------
    VideoFileStream
        Videofilestream object of the output video.

    """
    if not isinstance(vfs, bw.VideoFileStream):
        raise TypeError(f"`vfs` parameter not a VideoFileStream object.")

    if videofilepath is None:
        videofilepath = derive_filepath(vfs.filepath, 'results',
                                        suffix='.mp4')

    videoframes = vfs.iter_frames(startat=startat, nframes=nframes)
    frames = (coords.iter_frames(nchannels=3, value=(0, 0, 255))
              .add_weighted(0.8, videoframes, 0.7))

    if draw_framenumbers is not None:
        frames = frames.draw_framenumbers(org=draw_framenumbers)
    if draw_mean:
        centers = coords.get_coordmean()
        frames = frames.draw_circles(centers=centers, radius=6, 
                                     color=(255, 100, 0), thickness=2, 
                                     linetype=16, shift=0)
        # centers_lp = np.array(
        #     [np.convolve(centers[:, 0], np.ones(7) / 7, 'same'),
        #      np.convolve(centers[:, 1], np.ones(7) / 7, 'same')]).T
        # frames = frames.draw_circles(centers=centers_lp, radius=6, color=(100, 255, 0), thickness=2, linetype=16, shift=0)
    vfs = frames.tovideo(videofilepath, framerate=vfs.avgframerate, crf=crf, 
                         scale=scale)
    return vfs


"""This module provides convenient functions for plotting video-related data.

These are based on Matplotlib. If you want to display things automatically
in Jupyter, use `%matplotlib inline`

"""

import cv2 as cv
import matplotlib.pyplot as plt
import matplotlib.patches as patches


__all__ = ['imshow_frame']


def imshow_frame(frame, fig=None, ax=None, figsize=None, cmap=None,
                 draw_rectangle=None ):
    """Create an matplotlib image plot of the frame.

    Parameters
    ----------
    frame : numpy array image
        Video frame.
    fig : matplotlib Figure object, optional
        Provide if you already have a figure in which the frame should be
        plotted.
    ax : matplotlib Axes object, optional
        Provide if you already have an axes in which the frame should be
        plotted.
    figsize : (float, float), optional
        Width, height in inches. If not provided, defaults to
        rcParams["figure.figsize"] = [6.4, 4.8].
    draw_rectangle : (int, int, int, int), optional
        Draw a rectangle on image. h1, h2, w1, w2. Origin is left top.

    Returns
    -------
    fig, ax : matplotlib.figure.Figure, matplotlib.axes.Axes
        The matplotlib Figure and Axes objects.

    """
    if fig is None:
        w = 14
        h = frame.shape[1] * w / frame.shape[0]
        fig, ax = plt.subplots(figsize=(w,h))
    if ax is None:
        fig.add_axes()
    if cmap is None:
        if frame.ndim == 3:
            cmap = None
            frame = cv.cvtColor(frame, cv.COLOR_BGR2RGB)
        elif frame.ndim ==2:
            cmap = 'gray'
    ax.imshow(frame.astype('uint8'), cmap=cmap)
    if draw_rectangle is not None:
        h1,h2,w1,w2 = draw_rectangle
        x,y, width, height = w1, h1, w2-w1, h2-h1 # we convert to mtpll coords
        rect = patches.Rectangle((x,y), width, height, linewidth=1,
                                 edgecolor='r',
                                 facecolor='none')

        # Add the patch to the Axes
        ax.add_patch(rect)

    return fig, ax